"""
Plotting logic for :class:`Aggregator`. These plots illustrate the relationshp
between parameters and metrics from various vantage points.

Used by ./aggregate.py
"""
import ubelt as ub

try:
    from line_profiler import profile
except ImportError:
    profile = ub.identity


def build_plotter(agg, rois, plot_config):
    """
    Used by :class:`Aggregator`, which will generally immediately call
    :func:`ParamPlotter.plot_requested`.

    Returns:
        ParamPlotter
    """
    from geowatch.mlops.smart_global_helper import SMART_HELPER
    from geowatch.utils import util_kwplot
    from kwutil import Yaml

    build_special_columns(agg)
    agg.build()
    single_table = table = agg.table
    single_table = preprocess_table_for_seaborn(agg, table)

    MARK_DELIVERED = plot_config.get('mark_delivered', False)
    if MARK_DELIVERED:
        SMART_HELPER.mark_delivery(single_table)

    modifier: util_kwplot.LabelModifier = SMART_HELPER.label_modifier()

    config_label_mappings = plot_config.get('label_mappings', {})
    modifier.update(config_label_mappings)

    if rois is not None:
        agg.build_macro_tables(rois)
        macro_table = agg.region_to_tables[agg.primary_macro_region].copy()
        macro_table = preprocess_table_for_seaborn(agg, macro_table)

        if MARK_DELIVERED:
            SMART_HELPER.mark_delivery(macro_table)
        param_to_palette = SMART_HELPER.shared_palettes(macro_table)
        if 0:
            SMART_HELPER.mark_star_models(macro_table)
    else:
        macro_table = None
        param_to_palette = SMART_HELPER.shared_palettes(single_table)

    plot_dpath = plot_config.get('plot_dpath', None)
    if plot_dpath is None:
        from geowatch.mlops.aggregate import hash_regions
        if rois is not None:
            region_hash = hash_regions(rois)
        else:
            region_hash = 'allrois'
        plot_dpath = (agg.output_dpath / 'plots')
    else:
        plot_dpath = ub.Path(plot_dpath)

    macro_plot_dpath = (plot_dpath / (f'macro-plots-{region_hash}'))

    USE_EFFECTIVE_HACK = 1
    if USE_EFFECTIVE_HACK:
        # Relabel the resolved params to use the "effective-params"
        # instead.
        # print(f'agg.mappings={agg.mappings}')
        for col, lut in agg.mappings.items():
            resolved_col = 'resolved_' + col
            for c in [col, resolved_col]:
                for table in [macro_table, single_table]:
                    if table is not None:
                        if resolved_col in table:
                            new = table[resolved_col].apply(lambda x: lut.get(x, x))
                            table[resolved_col] = new

    vantage_points = plot_config.get('vantage_points', None)
    plotter = ParamPlotter(agg, vantage_points=vantage_points)

    plotter.plot_dpath = plot_dpath
    plotter.macro_plot_dpath = macro_plot_dpath

    # Maps paramter values to colors
    plotter.param_to_palette = param_to_palette

    # This is for remapping parameter values
    plotter.param_to_valmap = Yaml.coerce(plot_config.get('param_to_valmap', {}))
    plotter.param_to_valmap = plotter.param_to_valmap or {}

    # The modifier is for remapping parameter names
    plotter.modifier = modifier
    plotter.label_modifier = modifier

    # The main data
    plotter.macro_table = macro_table
    plotter.single_table = single_table

    plotter.rois = rois
    plotter.plot_config = plot_config
    plotter.roi_attr = 'region_id'
    return plotter


@profile
def build_all_param_plots(agg, rois, plot_config):
    """
    Main entry point for plotting results from an :class:`Aggregator`.
    """
    plotter = build_plotter(agg, rois, plot_config)
    plotter.plot_requested()


def build_special_columns(agg):
    from geowatch.utils import util_pandas
    resolved_params = util_pandas.DotDictDataFrame(agg.resolved_params)
    part1 = resolved_params.query_column('batch_size')
    if len(part1) > 1:
        # Disambiguate fit and pred batch size
        part1_ = [p for p in part1 if '_fit' in p]
        if len(part1_) == 1:
            part1 = part1_
    part2 = resolved_params.query_column('accumulate_grad_batches')
    prefix_to_batchsize = ub.group_items(part1, key=lambda x: x.rsplit('.', 2)[0])
    prefix_to_accumbatch = ub.group_items(part2, key=lambda x: x.rsplit('.', 2)[0])
    prefixes = set(prefix_to_batchsize) | set(prefix_to_accumbatch)

    for prefix in prefixes:
        cols1 = prefix_to_batchsize.get(prefix, None)
        cols2 = prefix_to_accumbatch.get(prefix, None)
        val_accum = 1
        val_bsize = resolved_params[cols1[0]]
        if cols2 is not None:
            assert len(cols2) == 1
            val_accum = resolved_params[cols2[0]].copy()
            val_accum[val_accum.isnull()] = 1
            val_accum[val_accum == 'None'] = 1
        val_effective_bsize = val_bsize * val_accum
        agg.table.loc[:, prefix + '.effective_batch_size'] = val_effective_bsize

    # agg.table['normalized_params.bas_pxl_fit.initializer.init'] = agg.table['resolved_params.bas_pxl_fit.initializer.init'].apply(lambda x: '/'.join(x.split('/')[-3:]))
    # HACK!!!
    # agg.table['normalized_params.bas_pxl_fit.initializer.init']


def preprocess_table_for_seaborn(agg, table):
    fillna_cols = table.columns.intersection(agg.resolved_params.columns.union(agg.resolved_params.columns))
    table.loc[:, fillna_cols] = table.loc[:, fillna_cols].fillna('None')
    table = table.applymap(lambda x: str(x) if isinstance(x, list) else x)

    from geowatch.utils import util_pandas
    from geowatch.mlops.smart_global_helper import SMART_HELPER
    table = util_pandas.DataFrame(table)
    channel_cols = table.match_columns('*.channels')
    if len(channel_cols):
        unique_channels = list(set(ub.flatten(table[channel_cols].value_counts().index)))
        try:
            unique_channels = sorted(unique_channels)
        except TypeError:
            ...
        channel_mapping = SMART_HELPER.custom_channel_relabel_mapping(unique_channels, coarsen=False)
        for col in channel_cols:
            table[col] = table[col].apply(channel_mapping.get)

    if 'resolved_params.bas_pxl_fit.initializer.init' in table:
        # SMART hack
        table['resolved_params.bas_pxl_fit.initializer.init'] = table['resolved_params.bas_pxl_fit.initializer.init'].apply(lambda x: '/'.join(x.split('/')[-3:]))
    # from geowatch.utils import util_pandas
    # table = util_pandas.DataFrame(table)
    # # table.match_columns('.channels')
    return table


class ParamPlotter:
    """
    Builds the scatter and box-and-whisker plots over different params.
    Working on cleaning this up
    """
    def __init__(plotter, agg, vantage_points=None):
        plotter.agg = agg

        # We will conduct analysis under serveral different vantage points
        if vantage_points is None:
            vantage_points = agg.default_vantage_points

        vantage_points = [Vantage(**vantage) for vantage in vantage_points]

        for vantage in vantage_points:
            pm = vantage['metric1'].split('.')[-1]
            sm = vantage['metric2'].split('.')[-1]
            name = f'{pm}-vs-{sm}'
            vantage['name'] = name
        plotter.vantage_points = vantage_points

    def plot_requested(plotter):
        """
        Simplified entry point
        """
        plot_config = plotter.plot_config
        plotter.plot_dpath.ensuredir()

        flag = plot_config.get('plot_resources', 'try')
        if flag:
            try:
                plotter.plot_resources()
            except Exception as ex:
                print(f'Error plotting resources ex={ex}')
                if flag != 'try':
                    raise

        if plot_config.get('plot_overviews', 1):
            plotter.plot_overviews()

        if plot_config.get('plot_params', 1):
            plotter.plot_params()

    def plot_resources(plotter):
        """
        Draw tables that summarize the resource usage of the experiments.
        """
        import rich
        from geowatch.utils import util_kwplot
        plotter.plot_dpath.ensuredir()
        agg = plotter.agg
        rich.print('[green] ### Plot Resources')
        # resource_summary_df = agg.resource_summary_table()
        resource_summary_df = agg.resource_summary_table_friendly()
        rich.print(resource_summary_df.to_string())
        resource_summary_df

        table = resource_summary_df
        table_title = 'resources'
        table_fpath = plotter.plot_dpath / f'{table_title}.png'
        table_style = table.style.set_caption(table_title)
        util_kwplot.dataframe_table(table_style, table_fpath, title=table_title)
        rich.print(f'Dpath: [link={plotter.plot_dpath}]{plotter.plot_dpath}[/link]')

        table_tex_fpath = plotter.plot_dpath / f'{table_title}.tex'
        tex = table.to_latex(index=False)
        tex = '\\begin{table*}[t]\n' + tex.rstrip('\n') + '\n\\end{table*}'
        print(tex)
        table_tex_fpath.write_text(tex)

    def plot_overviews(plotter):
        """
        Draw the overview for each vantage point.
        Draw tables that summarize the resource usage of the experiments.
        """
        from kwutil.util_progress import ProgressManager
        import rich
        pman = ProgressManager()

        rich.print('[green]### Plot Overviews')
        rich.print(f'Dpath: [link={plotter.plot_dpath}]{plotter.plot_dpath}[/link]')
        with pman:
            for vantage in pman.progiter(plotter.vantage_points, desc='plotting vantage overviews'):
                plotter.plot_vantage_per_region_overview(vantage)

                if plotter.macro_table is not None:
                    plotter.plot_vantage_macro_overview(vantage)

        rich.print(f'Dpath: [link={plotter.plot_dpath}]{plotter.plot_dpath}[/link]')

    def plot_params(plotter):
        from kwutil.util_progress import ProgressManager
        import rich
        pman = ProgressManager()

        rich.print('[green] ### Plot Params')
        if plotter.macro_table is None:
            rich.print(f'[red] Cannot plot params plotter.macro_table={plotter.macro_table}')
            return
        rich.print(f'Dpath: [link={plotter.macro_plot_dpath}]{plotter.macro_plot_dpath}[/link]')
        with pman:
            for vantage in pman.progiter(plotter.vantage_points, desc='plotting vantage params'):
                plotter.plot_vantage_params(vantage, pman=pman)
        rich.print(f'Dpath: [link={plotter.macro_plot_dpath}]{plotter.macro_plot_dpath}[/link]')

    def plot_vantage_per_region_overview(plotter, vantage):
        """
        Draw scatter plots and box plots that that distinguish each region with
        respect to a vantage point.

        Args:
            vantage (Dict):
                The primary and secondary metric to look at results with.
                This must have keys: metric1, metric2 and can optionally
                contain keys: scale1, scale2, objective1, and objective2.
        """
        from geowatch.utils import util_kwplot
        from geowatch.utils.util_kwplot import scatterplot_highlight
        import numpy as np
        import kwplot
        import kwimage
        import rich
        from geowatch.mlops.smart_global_helper import SMART_HELPER
        sns = kwplot.autosns()
        plt = kwplot.autoplt()  # NOQA
        kwplot.close_figures()

        rich.print('[white]### Plot Vantage Region Overview:')
        rich.print(f'[white] * {vantage}')

        agg = plotter.agg
        single_table = plotter.single_table

        name = vantage['name']
        y = vantage['metric1']
        x = vantage['metric2']
        yscale = vantage['scale1']
        xscale = vantage['scale2']
        main_metric = y
        roi_attr = plotter.roi_attr

        plotter.plot_dpath.ensuredir()
        finalize_figure = util_kwplot.FigureFinalizer(
            dpath=plotter.plot_dpath,
            size_inches=np.array([6.4, 4.8]) * 1.0,
        )
        fig = kwplot.figure(fnum=2, doclf=True)

        snskw = {}
        if roi_attr in plotter.param_to_palette:
            roi_to_color = util_kwplot.Palette.coerce(plotter.param_to_palette[roi_attr])
        else:
            unique_rois = single_table[roi_attr].unique()
            roi_to_color = util_kwplot.Palette.coerce(unique_rois)
        snskw['palette'] = roi_to_color

        s = plotter.plot_config.get('scatter.markersize', None)
        scatterkw = snskw.copy()
        if s is not None:
            scatterkw['s'] = s
        ax = sns.scatterplot(data=single_table, x=x, y=y, hue=roi_attr,
                             legend=False, **scatterkw)

        if plotter.plot_config.get('compare_sv_hack', False):
            # Hack to compare before/after SV
            if 'sv_poly_eval' in x.split('.'):
                plotter._add_sv_hack_lines(ax, single_table, x, y)
        if 'delivered_params' in single_table:
            val_to_color = SMART_HELPER.delivery_to_color
            if 0:
                kwplot.imshow(kwplot.make_legend_img(val_to_color, mode='star', dpi=300))
            scatterplot_highlight(data=single_table, x=x, y=y,
                                  highlight='delivered_params', ax=ax,
                                  color='group',
                                  size=300, val_to_color=val_to_color)
        ax.set_title(f'Per-Region Results (n={len(agg)})')
        ax.set_xscale(xscale)
        ax.set_yscale(yscale)
        plotter.modifier.relabel(ax, ticks=False)
        finalize_figure.finalize(fig, f'overview-{name}.png')
        rich.print(f'[green] made overview-{name}.png')

        try:
            macro_fig_final = finalize_figure.copy()
            macro_fig_final.update(size_inches=[12, 4], dpi=300)
            fig = kwplot.figure(fnum=90, doclf=True)

            region_order = plotter.plot_config.get('region_order', None)
            boxsns_kw = {}
            if region_order is not None:
                boxsns_kw['order'] = region_order
            boxsns_kw.update(snskw)

            util_kwplot.fix_seaborn_palette_issue(x=roi_attr, snskw=boxsns_kw)
            ax = sns.boxplot(data=single_table, x=roi_attr, y=main_metric, **boxsns_kw)
            ax.set_title(f'Per-Region Results (n={len(agg)})')
            param_histogram = single_table.groupby(roi_attr).size().to_dict()
            util_kwplot.LabelModifier({
                param_value: f'{param_value}\n(n={num})'
                for param_value, num in param_histogram.items()
            }).relabel_xticks(ax)
            plotter.modifier.relabel(ax, ticks=False)
            macro_fig_final.finalize(fig, f'overview-boxplot-{roi_attr}-vs-{main_metric}.png')

        except Exception as ex:
            rich.print(f'[red] warning, unable to plot overview-boxplot-region-vs-{main_metric}.png ex={ex}')
        else:
            rich.print(f'[green] made overview boxplot overview-boxplot-region-vs-{main_metric}.png')

        roi_legend = roi_to_color.make_legend_img(dpi=300)
        roi_legend_fpath = (plotter.plot_dpath / 'roi_legend.png')
        kwimage.imwrite(roi_legend_fpath, roi_legend)
        rich.print('[green] made roi_legend.png')

    def plot_vantage_macro_overview(plotter, vantage):
        """
        Draw a scatter plot that gives an overview of the requested macro table
        wrt to a metric vantage point.

        Args:
            vantage (Dict):
                The primary and secondary metric to look at results with.
                This must have keys: metric1, metric2 and can optionally
                contain keys: scale1, scale2, objective1, and objective2.
        """
        from geowatch.utils import util_kwplot
        from geowatch.utils.util_kwplot import scatterplot_highlight
        import kwplot
        import kwimage
        import numpy as np
        import rich
        from geowatch.mlops.smart_global_helper import SMART_HELPER

        rich.print('[white]### Plot Vantage Macro Overview:')
        rich.print(f'[white] * {vantage}')

        sns = kwplot.autosns()
        plt = kwplot.autoplt()  # NOQA
        kwplot.close_figures()

        rois = plotter.rois
        macro_table = plotter.macro_table
        roi_attr = plotter.roi_attr

        name = vantage['name']
        y = vantage['metric1']
        x = vantage['metric2']
        yscale = vantage['scale1']
        xscale = vantage['scale2']

        roi_finalizer = util_kwplot.FigureFinalizer(
            dpath=plotter.macro_plot_dpath,
            size_inches=np.array([6.4, 4.8]) * 1.0,
        )

        fig = kwplot.figure(fnum=3, doclf=True)
        ax = fig.gca()
        region_ids = macro_table[roi_attr].unique()
        assert len(region_ids) == 1
        macro_region_id = region_ids[0]
        palette = {
            macro_region_id: kwimage.Color('kitware_darkgray').as01()
        }

        s = plotter.plot_config.get('scatter.markersize', None)
        snskw = {}
        scatterkw = snskw.copy()
        if s is not None:
            scatterkw['s'] = s
        ax = sns.scatterplot(data=macro_table, x=x, y=y, hue=roi_attr, ax=ax,
                             palette=palette, **snskw)
        if plotter.plot_config.get('compare_sv_hack', False):
            # Hack to compare before/after SV
            if 'sv_poly_eval' in x.split('.'):
                plotter._add_sv_hack_lines(ax, macro_table, x, y)
        if 'is_star' in macro_table:
            scatterplot_highlight(
                data=macro_table, x=x, y=y, highlight='is_star', ax=ax,
                size=300)
        if 'delivered_params' in macro_table:
            import kwimage
            val_to_color = SMART_HELPER.delivery_to_color
            if 0:
                kwplot.imshow(kwplot.make_legend_img(val_to_color, mode='star', dpi=300))
            scatterplot_highlight(data=macro_table, x=x, y=y,
                                  highlight='delivered_params', ax=ax,
                                  color='group', size=300,
                                  val_to_color=val_to_color)

        from geowatch.utils.util_kwplot import TitleBuilder
        title_builder = TitleBuilder()
        title_builder.add_part(f'Results (n={len(macro_table)})')
        if [r for r in rois if r]:
            title_builder.ensure_newline()
            title_builder.add_part(f'Macro Analysis over {ub.urepr(rois, sv=1, nl=0)}')
        ax.set_title(title_builder.finalize())
        ax.set_xscale(xscale)
        ax.set_yscale(yscale)
        plotter.modifier.relabel(ax, ticks=False)
        roi_finalizer.finalize(fig, f'overview-macro_results-{name}.png')
        rich.print('[green] made overview-macro_results-{name}.png')

    def plot_vantage_params(plotter, vantage, pman=None, params_of_interest=None):
        """
        The main parameter inspection plots.

        A vantage point specifies the metrics to visualize and analyze.  Makes
        scatter plot, box plots, and attempts to draw legends useful for
        communicating results.

        The main per-paramter logic is in :func:`_plot_single_vantage_param`.

        Writes two folders in the "macro-plots-*" folder: params, and vantage.
        Both folders contain the same data, but it is laid out differently.

        1. The params folder contains a sub-folder for each parameter, and that
           folder contains plots for all vantage points.

        2. The "vantage" folder contains a sub-folder for each vantage point
           and all parameters plotted with that vantage point are plotted here.
           This folder will typically be symlinks into the above "params"
           folder.

        Args:
            vantage (Dict):
                The primary and secondary metric to look at results with.
                This must have keys: metric1, metric2 and can optionally
                contain keys: scale1, scale2, objective1, and objective2.

            pman (ProgressManager | None):
                a progress manager used to indicate sub-progress of plotting
        """
        import kwplot
        import rich
        from geowatch.utils import util_pandas
        rich.print('[white]### Plot Vantage Params:')
        rich.print(f'[white] * {vantage}')

        plt = kwplot.autoplt()  # NOQA
        kwplot.autosns()
        kwplot.close_figures()

        macro_table = plotter.macro_table

        main_metric = vantage['metric1']
        main_objective = vantage['objective1']
        metric_objectives = {main_metric: main_objective}

        from geowatch.mlops.smart_global_helper import SMART_HELPER
        blocklist = SMART_HELPER.VIZ_BLOCKLIST

        resolved_params = util_pandas.DotDictDataFrame(macro_table).subframe('resolved_params', drop_prefix=False)
        resolved_params['param_hashid'] = macro_table['param_hashid']
        valid_cols = resolved_params.columns.difference(blocklist)
        resolved_params = resolved_params[valid_cols]

        from kwutil.util_yaml import Yaml
        if params_of_interest is None:
            params_of_interest = Yaml.coerce(plotter.plot_config.get('params_of_interest', None))

        if params_of_interest is not None:
            chosen_params = params_of_interest
            params_of_interest = set(params_of_interest)
            valid_params_of_interest = list(resolved_params.columns.intersection(params_of_interest))
            missing = sorted(set(params_of_interest) - set(valid_params_of_interest))
            chosen_params = valid_params_of_interest
            if missing:
                rich.print('[yellow]WARNING: unknown params of interest!')
                rich.print('missing: {}'.format(ub.repr2(missing)))
                print('chosen_params = {}'.format(ub.urepr(chosen_params, nl=1)))
                suggest_did_you_mean(missing, resolved_params.columns)
        else:
            print('params_of_interest is unspecified, automatically choosing')

        # TODO: cleanup logic
        DO_STAT_ANALYSIS = plotter.plot_config.get('stats_ranking', False)
        if DO_STAT_ANALYSIS:
            ### Build param analysis
            from geowatch.utils import result_analysis
            metrics_table = util_pandas.DotDictDataFrame(macro_table).subframe('metrics', drop_prefix=False)
            results = {'params': resolved_params,
                       'metrics': metrics_table}
            # agg.primary_metric_cols)
            # TODO: params_of_interest in analysis
            analysis = result_analysis.ResultAnalysis(
                results, metrics=[main_metric], metric_objectives=metric_objectives)
            analysis.build()
            analysis.analysis()
            print('analysis.varied = {}'.format(ub.urepr(analysis.varied, nl=2)))
            ranked_stats = list(sorted(analysis.statistics, key=lambda x: x['anova_rank_p']))
            param_name_to_stats = {s['param_name']: s for s in ranked_stats}
            ranked_params = ub.oset(param_name_to_stats.keys())
            chosen_params = ranked_params
            if params_of_interest is not None:
                chosen_params = params_of_interest
        else:
            if params_of_interest is None:
                chosen_params = []
                for col in resolved_params.columns:
                    if len(macro_table[col].unique()) > 1:
                        chosen_params.append(col)
            param_name_to_stats = {}

        # ranked_params = ['bas_poly_eval.params.bas_pxl.package_fpath']
        if not len(chosen_params):
            print('Warning: no chosen params')
        else:
            print('chosen_params = {}'.format(ub.urepr(chosen_params, nl=1)))

        from kwutil.util_progress import ProgressManager
        owns_pman = 0

        results = []
        if pman is None:
            pman = ProgressManager()
            pman.__enter__()
            owns_pman = 1
        try:
            for rank, param_name in enumerate(pman.progiter(chosen_params, desc='plot param for ' + vantage['name'], verbose=3)):
                try:
                    drawn_rows = plotter._plot_single_vantage_param(
                        rank, macro_table, param_name, vantage,
                        params_of_interest, param_name_to_stats)
                    results.append(drawn_rows)
                except SkipPlot:
                    continue

            rich.print(f'Dpath: [link={plotter.macro_plot_dpath}]{plotter.macro_plot_dpath}[/link]')
            # agg0.analyze()
        finally:
            if owns_pman:
                pman.__exit__()
        return results

    def _plot_single_vantage_param(plotter, rank, macro_table, param_name,
                                   vantage, params_of_interest,
                                   param_name_to_stats):
        """
        Inner loop for :func:`ParamPlotter.plot_vantage_params`,
        todo: reduce arguments
        """
        import rich
        import kwplot
        import kwarray
        import pandas as pd
        from kwcoco.metrics.drawing import concice_si_display
        from geowatch.utils import util_kwplot
        from geowatch.utils.util_kwplot import scatterplot_highlight
        import numpy as np
        import seaborn as sns

        rich.print(f'param_name = {ub.urepr(param_name, nl=1)}')

        param_group_dpath = plotter.macro_plot_dpath / 'params'
        param_to_palette = plotter.param_to_palette
        vantage_dpath = ((plotter.macro_plot_dpath / 'vantage' / vantage['name']).ensuredir()).resolve()
        vantage_flat_dpath = (vantage_dpath / '_flat').ensuredir()
        finalize_figure = util_kwplot.FigureFinalizer(
            size_inches=np.array([6.4, 4.8]) * 1.0,
        )

        modifier = plotter.modifier

        rois = plotter.rois
        main_metric = y = vantage['metric1']
        secondary_metric = x = vantage['metric2']
        yscale = vantage['scale1']
        xscale = vantage['scale2']

        stats = param_name_to_stats.get(param_name, {})
        # stats['moments']
        anova_rank_p = stats.get('anova_rank_p', None)
        # param_name = stats['param_name']

        # if 0:
        #     # macro_table[param_name] = macro_table[param_name].apply(lambda x: config_label_mappings.get(x, x))
        #     param_to_palette[param_name] = SMART_HELPER.make_param_palette(macro_table[param_name].unique())

        try:
            macro_table = macro_table.sort_values(param_name)
        except Exception as ex:
            rich.print(f'[yellow] warning, unable to sort values by {param_name} ex={ex}')

        # Number of samples we have for each value of this parameter
        raw_param_histogram = ub.udict(macro_table.groupby(param_name).size().to_dict())
        orig_unique_params = ub.udict()
        param_histogram = ub.udict()
        for k, v in raw_param_histogram.items():
            str_key = str(k)
            if str_key in param_histogram:
                param_histogram[str_key] += v
                orig_unique_params[str_key].append(k)
            else:
                orig_unique_params[str_key] = [k]
                param_histogram[str_key] = v

        sub_macro_table = macro_table
        num_macro_rows = len(macro_table)

        min_variations = plotter.plot_config.get('min_variations', 1)  # if less than this number of groups, skip the plot
        max_variations = plotter.plot_config.get('max_variations', float('inf'))  # if less than this number of groups, skip the plot
        min_support = plotter.plot_config.get('min_support', 1)  # minimum amount of values in a group, otherwise remove that group

        explicitly_requested = (params_of_interest is not None and param_name in params_of_interest)

        if min_support > 1 and not explicitly_requested:
            ignore_params = [k for k, v in param_histogram.items() if v < min_support]
            param_histogram = ub.udict(param_histogram) - set(ignore_params)
            row_is_ignored = kwarray.isect_flags(macro_table[param_name], ignore_params)
            sub_macro_table = macro_table[~row_is_ignored]

        num_filtered_rows = num_macro_rows - len(sub_macro_table)
        n_support_stats = kwarray.stats_dict(list(param_histogram.values()), quantile=False, median=True)
        unique_val_types = list(map(type, ub.flatten(orig_unique_params.take(param_histogram))))
        value_type_hist = ub.dict_hist(unique_val_types)
        column_dtype = str(sub_macro_table[param_name].dtype)
        param_summary = {
            'param_name': param_name,
            'num_rows': len(sub_macro_table),
            'num_filtered_rows': num_filtered_rows,
            # Does this need a better name? It is showing the distribution of
            # support for the different groupings.
            'n_support_stats': n_support_stats,
            'column_dtype': column_dtype,
            'value_type_hist': value_type_hist,
        }
        rich.print(f'param_summary = {ub.urepr(param_summary, nl=2)}')

        if len(param_histogram) < min_variations:
            msg = f'skip plot for param_name={param_name} due to min_variations'
            rich.print(f'[red] {msg}')
            raise SkipPlot(msg)

        if len(param_histogram) > max_variations:
            msg = f'skip plot for param_name={param_name} due to max_variations'
            rich.print(f'[red] {msg}')
            raise SkipPlot(msg)

        if 1:
            # HACK: if the parameter we are plotting doesn't have a palette, make one
            # probably want to build palettes outside of this function?
            if param_name not in param_to_palette:
                # We may want to sort the input here?
                param_to_palette[param_name] = util_kwplot.Palette.coerce(macro_table[param_name].unique())

        snskw = {}
        if param_name in param_to_palette:
            snskw['palette'] = param_to_palette[param_name]

        s = plotter.plot_config.get('scatter.markersize', None)
        scatterkw = snskw.copy()
        if s is not None:
            scatterkw['s'] = s

        from geowatch.utils.util_kwplot import TitleBuilder
        title_builder = TitleBuilder()
        title_builder.add_part(f'Results (n={len(sub_macro_table)})')
        if [r for r in rois if r]:
            title_builder.ensure_newline()
            title_builder.add_part(f'Macro Analysis over {ub.urepr(rois, sv=1, nl=0)}')
        if anova_rank_p is not None:
            title_builder.ensure_newline()
            title_builder.append(f'Effect of {param_name}: anova_rank_p={concice_si_display(anova_rank_p)}')
        header_text = title_builder.finalize()

        param_valname_map, had_value_remap = shrink_param_names(param_name, list(param_histogram))
        print(f'had_value_remap={had_value_remap}')

        if param_name in plotter.param_to_valmap:
            had_value_remap = True
            # User can overload the mappign of the parameter value names
            param_valname_map.update(plotter.param_to_valmap[param_name])
            ...

        # Mapper for the legend
        if had_value_remap:
            freq_mapper_scatter = util_kwplot.LabelModifier({
                param_value: f'{param_value}\n{param_valname_map[param_value]} (n={num})'
                for param_value, num in param_histogram.items()
            })
        else:
            freq_mapper_scatter = util_kwplot.LabelModifier({
                param_value: f'{param_value}\n(n={num})'
                for param_value, num in param_histogram.items()
            })

        freq_mapper_box = util_kwplot.LabelModifier({
            param_value: f'{param_valname_map[param_value]}\n(n={num})'
            for param_value, num in param_histogram.items()
        })

        vantage_prefix = f'macro_results_{rank:03d}_{param_name}'
        param_prefix = f'macro_results_{param_name}'
        param_metric_prefix = f'{param_prefix}_{main_metric}'
        param_metric2_prefix = f'{param_prefix}_{main_metric}_vs_{secondary_metric}'

        drawn_rows = []

        # SCATTER
        fig = kwplot.figure(fnum=4, doclf=True)
        # Scatter with legend
        ax = sns.scatterplot(data=sub_macro_table, x=x, y=y,
                             hue=param_name, legend=True, **scatterkw)
        ax.set_title(header_text)
        if 'is_star' in sub_macro_table:
            scatterplot_highlight(data=sub_macro_table, x=x, y=y, highlight='is_star', ax=ax, size=300)

        if plotter.plot_config.get('compare_sv_hack', False):
            # Hack to compare before/after SV
            if 'sv_poly_eval' in x.split('.'):
                plotter._add_sv_hack_lines(ax, sub_macro_table, x, y)

        ax.set_xscale(xscale)
        ax.set_yscale(yscale)
        modifier.relabel(ax, ticks=False)

        param_dpath = (param_group_dpath / param_name).ensuredir().resolve()
        param_dpath_flat = (param_dpath / '_flat').ensuredir()

        param_fpath = param_dpath_flat / f'{param_metric2_prefix}_PLT01_scatter_legend.png'
        finalize_figure.finalize(fig, param_fpath)
        rich.print(f'[green] wrote {param_fpath.name}')
        drawn_rows.append({
            'type': 'vantage_metric_plot',
            'plot_type': 'scatter_legend',
            'param_fpath': param_fpath,
            'suffix': 'PLT01_scatter_legend.png',
            'prefix': param_metric2_prefix,
        })

        force_regen = True

        # Scatter legend (doesnt care about the vantage)
        param_fpath = param_dpath_flat / f'{param_prefix}_PLT03_scatter_onlylegend.png'
        try:
            if not param_fpath.exists() or force_regen:
                legend_ax = util_kwplot.extract_legend(ax)
                freq_mapper_scatter.relabel(legend_ax, ticks=False)
                finalize_figure.finalize(legend_ax.figure, param_fpath)
                drawn_rows.append({
                    'type': 'main_param_legend',
                    'param_fpath': param_fpath,
                    'suffix': 'PLT03_scatter_onlylegend.png',
                    'prefix': param_prefix,
                })
        except RuntimeError:
            rich.print(f'[red] failed to write {param_fpath.name}')
        else:
            rich.print(f'[green] wrote {param_fpath.name}')
            legend = ax.get_legend()
            if legend is not None:
                legend.remove()
        # Scatter without legend
        param_fpath = param_dpath_flat / f'{param_metric2_prefix}_PLT02_scatter_nolegend.png'
        finalize_figure.finalize(fig, param_fpath)
        rich.print(f'[green] wrote {param_fpath.name}')
        drawn_rows.append({
            'type': 'vantage_metric_plot',
            'plot_type': 'scatter_nolegend',
            'param_fpath': param_fpath,
            'suffix': 'PLT02_scatter_nolegend.png',
            'prefix': param_metric2_prefix,
        })

        # BOX
        param_fpath = param_dpath_flat / f'{param_metric_prefix}_PLT04_box.png'
        if not param_fpath.exists() or force_regen:
            fig = kwplot.figure(fnum=5, doclf=True)
            util_kwplot.fix_seaborn_palette_issue(x=param_name, snskw=snskw)
            try:
                ax = sns.boxplot(data=sub_macro_table, x=param_name, y=y,
                                 legend=False, **snskw)
            except Exception as box_ex:
                box_ex = box_ex
                rich.print(ub.codeblock(
                    f'''
                    [red]Error with boxplot

                    x={param_name}
                    y={y}

                    ex={box_ex}
                    '''))
            else:
                freq_mapper_box.relabel_xticks(ax)
                ax.set_title(header_text)
                modifier.relabel(ax, ticks=False)
                modifier.relabel_xticks(ax)
                finalize_figure.finalize(fig, param_fpath)
                rich.print(f'[green] wrote {param_fpath.name}')
                drawn_rows.append({
                    'plot_type': 'box',
                    'type': 'main_metric_plot',
                    'param_fpath': param_fpath,
                    'suffix': 'PLT04_box.png',
                    'prefix': param_metric_prefix,
                })

        # Output dataframe table for larger legend
        # Varied value table (doesn't care about the vantage)
        param_fpath = param_dpath_flat / f'{param_prefix}_PLT05_table.png'
        if not param_fpath.exists() or force_regen:
            # TODO: don't check for existence, check if we generated it
            # previously in this sessions (we want this to regenerate if
            # we modify code).
            param_code_lut = []
            for old_name, new_name in param_valname_map.items():
                param_code_lut.append({
                    'code': new_name,
                    'value': old_name,
                    'num': param_histogram[old_name],
                })
            param_code_lut = pd.DataFrame(param_code_lut, columns=['code', 'value', 'num'])
            if not had_value_remap:
                param_code_lut = param_code_lut.drop('code', axis=1)
            param_title = 'Key: ' + modifier._modify_text(param_name)
            lut_style = param_code_lut.style.set_caption(param_title)
            util_kwplot.dataframe_table(lut_style, param_fpath, title=param_title)
            rich.print(f'[green] wrote {param_fpath.name}')
            drawn_rows.append({
                'type': 'main_param_legend',
                'param_fpath': param_fpath,
                'suffix': 'PLT05_table.png',
                'prefix': param_prefix,
            })

        # Symlink to other organization structures
        for row in drawn_rows:
            param_fpath = row['param_fpath']
            suffix = row['suffix']
            prefix = row['prefix']
            vantage_fpath = vantage_flat_dpath / f'{vantage_prefix}_{suffix}'
            ub.symlink(real_path=param_fpath, link_path=vantage_fpath, overwrite=True)

            if row.get('plot_type') is not None:
                param_plot_type_dpath = (param_dpath / row['plot_type']).ensuredir()
                link_fpath = param_plot_type_dpath / f'{prefix}_{suffix}'
                ub.symlink(real_path=param_fpath, link_path=link_fpath, overwrite=True)

                vantage_plot_type_dpath = (vantage_dpath / row['plot_type']).ensuredir()
                link_fpath = vantage_plot_type_dpath / f'{vantage_prefix}_{suffix}'
                ub.symlink(real_path=param_fpath, link_path=link_fpath, overwrite=True)

            if row['type'] == 'main_param_legend':
                link_fpath = param_dpath / f'{prefix}_{suffix}'
                ub.symlink(real_path=param_fpath, link_path=link_fpath, overwrite=True)
            if row['type'] == 'vantage_metric_plot':
                ...
            if row['type'] == 'main_metric_plot':
                ...
        # ub.symlink(real_path=param_fpath, link_path=vantage_fpath, overwrite=True)
        return drawn_rows

    def _add_sv_hack_lines(plotter, ax, table, x, y):
        import matplotlib as mpl

        def add_arrows_to_lines(line_collection, position=None, direction='right', size=15, color=None):
            """
            add an arrow to a line.

            line:       Line2D object
            position:   x-position of the arrow. If None, mean of xdata is taken
            direction:  'left' or 'right'
            size:       size of the arrow in fontsize points
            color:      if None, line color is taken.

            References:
                .. [SO34017866] https://stackoverflow.com/questions/34017866/arrow-on-a-line-plot-with-matplotlib
            """
            if color is None:
                color = line_collection.get_color()

            for segment in line_collection.get_segments():
                xdata = segment[:, 0]
                ydata = segment[:, 1]

                if position is None:
                    position = xdata.mean()
                # find closest index
                import numpy as np
                start_ind = np.argmin(np.absolute(xdata - position))
                start_ind = 0
                if direction == 'right':
                    end_ind = start_ind + 1
                else:
                    end_ind = start_ind - 1

                line_collection.axes.annotate(
                    '',
                    xytext=(xdata[start_ind], ydata[start_ind]),
                    xy=(xdata[end_ind], ydata[end_ind]),
                    arrowprops=dict(arrowstyle="->", color=color),
                    size=size, zorder=0,
                )
        # Hack to compare before/after SV
        # import matplotlib as mpl
        # ax = sns.scatterplot(data=single_table, x=x, y=y, hue='region_id', legend=False)
        # sns.scatterplot(data=single_table, x=x_prev, y=y_prev, ax=ax, legend=False)
        if 'sv_poly_eval' in x.split('.'):
            print('SV HACK!!!')
            x_prev = x.replace('sv_poly_eval', 'bas_poly_eval')
            y_prev = y.replace('sv_poly_eval', 'bas_poly_eval')

            # xy1 = table[[x_prev, y_prev]].values
            # xy2 = table[[x, y]].values
            # uv = xy2 - xy1
            # ax.quiver(xy1.T[0], xy1.T[1], uv.T[0], uv.T[1])

            segments = []
            # patches = []
            for x1, y1, x2, y2 in table[[x_prev, y_prev, x, y]].values:
                segments.append([(x1, y1), (x2, y2)])
                # patch = mpl.patches.FancyArrow(
                #     x1, y1, x2 - x1, y2 - y1,
                #     width=0.001, length_includes_head=True,
                #     head_width=0.001,
                #     head_length=0.001,
                # )
                # patches.append(patch)
                ...
            # collection = mpl.collections.PatchCollection(patches)
            # ax.add_collection(collection)
            line_collection = mpl.collections.LineCollection(segments, color='blue', alpha=0.5, linewidths=1)
            ax.add_collection(line_collection)
            add_arrows_to_lines(line_collection)
            # pts1 = [s[0] for s in segments]
            # pts2 = [s[1] for s in segments]
            # ax.plot(*zip(*pts1), 'rx', label='before SV')
            # ax.plot(*zip(*pts2), 'bo', label='after SV')


def edit_distance(string1, string2):
    """
    Edit distance algorithm. String1 and string2 can be either
    strings or lists of strings

    Args:
        string1 (str | List[str]):
        string2 (str | List[str]):

    Requirements:
        pip install python-Levenshtein

    Returns:
        float | List[float] | List[List[float]]

    Example:
        >>> # xdoctest: +REQUIRES(module:Levenshtein)
        >>> string1 = 'hello world'
        >>> string2 = ['goodbye world', 'rofl', 'hello', 'world', 'lowo']
        >>> edit_distance(['hello', 'one'], ['goodbye', 'two'])
        >>> edit_distance('hello', ['goodbye', 'two'])
        >>> edit_distance(['hello', 'one'], 'goodbye')
        >>> edit_distance('hello', 'goodbye')
        >>> distmat = edit_distance(string1, string2)
        >>> result = ('distmat = %s' % (ub.repr2(distmat),))
        >>> print(result)
        >>> [7, 9, 6, 6, 7]
    """

    import Levenshtein
    isiter1 = ub.iterable(string1)
    isiter2 = ub.iterable(string2)
    strs1 = string1 if isiter1 else [string1]
    strs2 = string2 if isiter2 else [string2]
    distmat = [
        [Levenshtein.distance(str1, str2) for str2 in strs2]
        for str1 in strs1
    ]
    # broadcast
    if not isiter2:
        distmat = [row[0] for row in distmat]
    if not isiter1:
        distmat = distmat[0]
    return distmat


def suggest_did_you_mean(invalid_options, valid_choices):
    """
    Args:
        missing (List[str]): the invalid options the user chose
        valid_choices (List[str]): the valid available options

    Example:
        >>> from geowatch.mlops.aggregate_plots import *  # NOQA
        >>> invalid_options = ['fuber', 'yam', 'spam']
        >>> valid_choices = ['spam', 'ham', 'foobar']
        >>> suggest_did_you_mean(invalid_options, valid_choices)
    """
    import numpy as np
    import rich
    try:
        distances = np.array(edit_distance(invalid_options, valid_choices))
        for got, dists in zip(invalid_options, distances):
            min_idx = dists.argmin()
            if dists[min_idx] != 0:
                alternative = valid_choices[min_idx]
                rich.print(f'[yellow] Got: {got}. Did you mean {alternative}?')
    except ImportError:
        rich.print('[yellow] Warning: unable to suggest existing alternatives')


def shrink_param_names(param_name, param_values, text_len_thresh=20):
    param_labels = [str(p) for p in param_values]
    text_label_size = len(''.join(param_labels))
    if text_label_size > text_len_thresh:
        had_value_remap = True
        # Param names are too long. need to map parameter names to codes.
        param_valname_map = {}
        prefixchar = param_name.split('.')[-1][0].upper()
        for idx, value in enumerate(sorted(param_values)):
            old_name = str(value)
            new_name = f'{prefixchar}{idx:02d}'
            param_valname_map[old_name] = new_name
    else:
        had_value_remap = False
        param_valname_map = ub.dzip(param_labels, param_labels)
    return param_valname_map, had_value_remap


class SkipPlot(Exception):
    ...


class Vantage2(dict):
    """
    The primary and secondary metric to look at results with.
    This must have keys: metric1, metric2 and can optionally
    contain keys: scale1, scale2, objective1, and objective2.
    """

    @property
    def name(self):
        return self['name']

    @property
    def metric1(self):
        return self['metric1']

    @property
    def metric2(self):
        return self['metric2']

    @property
    def scale1(self):
        return self.get('scale1', 'linear')

    @property
    def scale2(self):
        return self.get('scale2', 'linear')


from dataclasses import dataclass  # NOQA


@dataclass
class Vantage:
    """
    The primary and secondary metric to look at results with.
    This must have keys: metric1, metric2 and can optionally
    contain keys: scale1, scale2, objective1, and objective2.

    Example:
        >>> from geowatch.mlops.aggregate_plots import *  # NOQA
        >>> Vantage('metrics.ppv', 'metrics.tpr')
    """
    metric1: str
    metric2: str
    scale1: str = 'linear'
    scale2: str = 'linear'
    objective1: str = 'maximize'
    objective2: str = 'maximize'
    name: str = None

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)



r"""
Writing cold kwcoco script after predict.py

CommandLine:

    #######################
    ### FULL REGION TEST-V1
    #######################

    DATA_DVC_DPATH=$(geowatch_dvc --tags=phase2_data --hardware="auto")
    EXPT_DVC_DPATH=$(geowatch_dvc --tags=phase2_expt --hardware="auto")
    python -m geowatch.tasks.cold.writing_kwcoco \
        --coco_fpath="$DATA_DVC_DPATH/Drop6/imgonly-KR_R001.kwcoco.json" \
        --combined_coco_fpath="$DATA_DVC_DPATH/Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip" \
        --out_dpath="$DATA_DVC_DPATH/Drop6-MeanYear10GSD-V2/_pycold_combine_V1" \
        --mod_coco_fpath="$DATA_DVC_DPATH/Drop6-MeanYear10GSD-V2/imgonly_KR_R001_cold-V1.kwcoco.zip" \
        --method='COLD' \
        --timestamp=False \
        --combine=True \
        --resolution='10GSD' \
        --workermode='serial' \
        --workers=0

    kwcoco stats "$DATA_DVC_DPATH"/Drop6-MeanYear10GSD-V2/imgonly_KR_R001_cold-V1.kwcoco.zip
    geowatch stats "$DATA_DVC_DPATH"/Drop6-MeanYear10GSD-V2/imgonly_KR_R001_cold-V1.kwcoco.zip
    kwcoco validate "$DATA_DVC_DPATH"/Drop6-MeanYear10GSD-V2/imgonly_KR_R001_cold-V1.kwcoco.zip

    geowatch visualize \
        "$DATA_DVC_DPATH"/Drop6-MeanYear10GSD-V2/imgonly_KR_R001_cold-V1.kwcoco.zip \
        --channels="L8:(red|green|blue,red_COLD_a1|green_COLD_a1|blue_COLD_a1,red_COLD_cv|green_COLD_cv|blue_COLD_cv,red_COLD_rmse|green_COLD_rmse|blue_COLD_rmse)" \
        --exclude_sensors=WV,PD,S2 \
        --smart=True

    #######################
    ### FULL REGION TEST V2
    #######################

    DATA_DVC_DPATH=$(geowatch_dvc --tags=phase2_data --hardware="auto")
    EXPT_DVC_DPATH=$(geowatch_dvc --tags=phase2_expt --hardware="auto")
    python -m geowatch.tasks.cold.writing_kwcoco \
        --coco_fpath="$DATA_DVC_DPATH/Drop6/imgonly-KR_R001.kwcoco.json" \
        --out_dpath="$DATA_DVC_DPATH//Drop6/_pycold_combine_V2" \
        --mod_coco_fpath="$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip" \
        --method='COLD' \
        --timestamp=False \
        --combine=False \
        --resolution='10GSD' \
        --workermode='serial' \
        --workers=0

    kwcoco stats "$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip"
    geowatch stats "$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip"
    kwcoco validate "$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip"

    DATA_DVC_DPATH=$(geowatch_dvc --tags=phase2_data --hardware="auto")
    geowatch visualize \
        "$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip" \
        --channels="L8:(red|green|blue,red_COLD_a1|green_COLD_a1|blue_COLD_a1,red_COLD_cv|green_COLD_cv|blue_COLD_cv,red_COLD_rmse|green_COLD_rmse|blue_COLD_rmse)" \
        --exclude_sensors=WV,PD,S2 \
        --smart=True

"""
import scriptconfig as scfg
import ubelt as ub
import os
import json
import logging
import kwcoco
import kwimage
import pandas as pd

logger = logging.getLogger(__name__)


try:
    from line_profiler import profile
except ImportError:
    profile = ub.identity


class WriteColdCocoConfig(scfg.DataConfig):
    """
    The docstring will be the description in the CLI help
    """

    coco_fpath = scfg.Value(None, position=1, help=ub.paragraph(
        '''
        a path to a file to input kwcoco file (to predict on)
        '''))
    combined_coco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        a path to a file to combined input kwcoco file (to merge with)
        '''))
    mod_coco_fpath = scfg.Value(None, help='file path for modified output coco json')
    out_dpath = scfg.Value(None, help='output directory for the output. If unspecified uses the output kwcoco bundle')
    method = scfg.Value('COLD', choices=['COLD', 'HybridCOLD', 'OBCOLD'], help='type of cold algorithms')
    timestamp = scfg.Value(False, help='True: exporting cold result by timestamp, False: exporting cold result by year, Default is False')
    combine = scfg.Value(True, help='for temporal combined mode, Default is True')
    track_emissions = scfg.Value(True, help='if True use codecarbon for emission tracking')
    resolution = scfg.Value('30GSD', help='if specified then data is processed at this resolution')
    workers = scfg.Value(8, help='total number of workers')
    workermode = scfg.Value('process', help='Can be process, serial, or thread')


@profile
def cold_writing_kwcoco_main(cmdline=1, **kwargs):
    """
    Args:
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m geowatch.tasks.cold.writing_kwcoco --help
        TEST_COLD=1 xdoctest -m geowatch.tasks.cold.writing_kwcoco cold_writing_kwcoco_main

     Example:
        >>> # xdoctest: +REQUIRES(env:TEST_COLD)
        >>> from geowatch.tasks.cold.writing_kwcoco import cold_writing_kwcoco_main
        >>> from geowatch.tasks.cold.writing_kwcoco import *
        >>> kwargs= dict(
        >>>   coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6/imgonly-KR_R001.kwcoco.json'),
        >>>   combined_coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip'),
        >>>   mod_coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6/imgonly-KR_R001.kwcoco.testing_comFalse.json'),
        >>>   out_dpath = ub.Path.appdir('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/_pycold_combine1'),
        >>>   method = 'COLD',
        >>>   timestamp = False,
        >>>   combine = False,
        >>>   resolution = '10GSD',
        >>>   workermode = 'process',
        >>> )
        >>> cmdline=0
        >>> cold_writing_kwcoco_main(cmdline, **kwargs)
    """
    #NOTE: This script doesn't consider timestamp = True
    config = WriteColdCocoConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = {}'.format(ub.urepr(config, nl=1)))

    from geowatch.utils import process_context
    from kwutil import util_json
    resolved_config = config.to_dict()
    resolved_config = util_json.ensure_json_serializable(resolved_config)

    proc_context = process_context.ProcessContext(
        name='geowatch.tasks.cold.writing_kwcoco',
        type='process',
        config=resolved_config,
        track_emissions=config['track_emissions'],
    )

    # Assign variables
    coco_fpath = ub.Path(config['coco_fpath'])
    combine = config['combine']

    if config['combined_coco_fpath'] is not None:
        combined_coco_fpath = ub.Path(config['combined_coco_fpath'])
    else:
        if combine:
            raise ValueError('Must specify combined_coco_fpath if combine is True')
        combined_coco_fpath = None

    if config['out_dpath'] is None:
        config['out_dpath'] = coco_fpath.parent

    mod_coco_fpath = ub.Path(config['mod_coco_fpath'])
    out_dpath = ub.Path(config['out_dpath']).ensuredir()
    method = config['method']
    resolution = config['resolution']
    metadata = read_json_metadata(out_dpath)
    region_id = metadata['region_id']
    cold_feat_path = out_dpath / 'reccg' / region_id / 'cold_feature'

    coef_names = ['cv', 'rmse', 'a0', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'c1']
    band_names = [0, 1, 2, 3, 4, 5]

    BAND_INFO = {0: 'blue',
                 1: 'green',
                 2: 'red',
                 3: 'nir',
                 4: 'swir16',
                 5: 'swir22'}

    proc_context.start()
    proc_context.add_disk_info(out_dpath)

    logger.info('Starting adding new asset to kwcoco json')

    asset_w = metadata['video_w']
    asset_h = metadata['video_h']

    if combine:
        combined_coco_dset = kwcoco.CocoDataset(combined_coco_fpath)
        coco_dset = kwcoco.CocoDataset(coco_fpath)

        for image_id in combined_coco_dset.images():
            combined_coco_image: kwcoco.CocoImage = combined_coco_dset.coco_image(image_id)
            coco_image: kwcoco.CocoImage = coco_dset.coco_image(image_id)
            image_name = coco_image.img['name']

            for band_name in band_names:
                for coef in coef_names:
                    band = BAND_INFO[band_name]
                    cold_feat_fpath = cold_feat_path / f'{image_name}_{band}_{method}_{coef}.tif'
                    if cold_feat_fpath.exists():
                        channels = kwcoco.ChannelSpec.coerce(f'{band}_{method}_{coef}')
                        # COLD output was wrote based on transform information of
                        # coco_dset, so it aligned to a scaled video space.
                        warp_img_from_vid = combined_coco_image.warp_img_from_vid

                        if resolution is None:
                            scale_asset_from_vid = (1., 1.)
                        else:
                            scale_asset_from_vid = combined_coco_image._scalefactor_for_resolution(
                                space='video', resolution=resolution)
                        warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)
                        warp_vid_from_asset = warp_asset_from_vid.inv()
                        warp_img_from_asset = warp_img_from_vid @ warp_vid_from_asset

                        # Use the CocoImage helper which will augment the coco dictionary with
                        # your information.
                        combined_coco_image.add_asset(
                            file_name=cold_feat_fpath,
                            channels=channels,
                            width=asset_w,
                            height=asset_h,
                            warp_aux_to_img=warp_img_from_asset)
                        # logger.info(f'Added to the asset {cold_feat_fpath}')
    else:
        coco_dset = kwcoco.CocoDataset(coco_fpath)

        # Get ordinal day list
        block_folder = out_dpath / 'stacked' / region_id / 'block_x1_y1'

        # if timestamp:
        meta_files = [m for m in os.listdir(block_folder) if m.endswith('.json')]

        # sort image files by ordinal dates
        img_dates = []
        img_names = []

        # read metadata and get img_name list of first date of each year
        for meta in meta_files:
            meta_config = json.loads((block_folder / meta).read_text())
            ordinal_date = meta_config['ordinal_date']
            img_name = meta_config['image_name']
            img_dates.append(ordinal_date)
            img_names.append(img_name)

        img_dates = sorted(img_dates)
        img_names = sorted(img_names)

        # Get only the first ordinal date of each year
        first_ordinal_dates = []
        first_img_names = []
        last_year = None
        for ordinal_day, img_name in zip(img_dates, img_names):
            year = pd.Timestamp.fromordinal(ordinal_day).year
            if year != last_year:
                first_ordinal_dates.append(ordinal_day)
                first_img_names.append(img_name)
                last_year = year

        img_names = first_img_names

        for image_id in coco_dset.images():
            # Create a CocoImage object for each image.
            coco_image: kwcoco.CocoImage = coco_dset.coco_image(image_id)
            image_name = coco_image.img['name']
            if image_name in img_names:
                for band_name in band_names:
                    for coef in coef_names:
                        band = BAND_INFO[band_name]
                        cold_feat_fpath = cold_feat_path / f'{image_name}_{band}_{method}_{coef}.tif'
                        if cold_feat_fpath.exists():
                            channels = kwcoco.ChannelSpec.coerce(f'{band}_{method}_{coef}')

                            # COLD output was wrote based on transform information of
                            # coco_dset, so it aligned to a scaled video space.
                            warp_img_from_vid = coco_image.warp_img_from_vid

                            if resolution is None:
                                scale_asset_from_vid = (1., 1.)
                            else:
                                scale_asset_from_vid = coco_image._scalefactor_for_resolution(
                                    space='video', resolution=resolution)
                            warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)
                            warp_vid_from_asset = warp_asset_from_vid.inv()
                            warp_img_from_asset = warp_img_from_vid @ warp_vid_from_asset

                            # Use the CocoImage helper which will augment the
                            # coco dictionary with your information.
                            coco_image.add_asset(os.fspath(cold_feat_fpath),
                                                 channels=channels,
                                                 width=asset_w, height=asset_h,
                                                 warp_aux_to_img=warp_img_from_asset)

                            logger.info(f'Added to the asset {cold_feat_fpath}')

    if proc_context is not None:
        context_info = proc_context.stop()
        coco_dset.dataset['info'].append(context_info)

    # Write a modified kwcoco.json file
    logger.info(f'Writing kwcoco file to: {mod_coco_fpath}')
    if combine:
        combined_coco_dset.fpath = mod_coco_fpath
        combined_coco_dset._ensure_json_serializable()
        combined_coco_dset.dump()
    else:
        coco_dset.fpath = mod_coco_fpath
        coco_dset._ensure_json_serializable()
        coco_dset.dump()
    logger.info(f'Finished writing kwcoco file to: {mod_coco_fpath}')


def read_json_metadata(folder_path):
    stacked_path = folder_path / 'stacked'
    for root, dirs, files in os.walk(stacked_path):
        for file in files:
            if file.endswith(".json"):
                json_path = os.path.join(root, file)

                with open(json_path, "r") as f:
                    metadata = json.load(f)
                    return metadata


if __name__ == '__main__':
    cold_writing_kwcoco_main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tensorflow code for working with object instances in segmentation."""

from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union

import tensorflow as tf


def instances_without_ignore_categories(panoptic_labels: tf.Tensor,
                                        ignore_categories: Union[tf.Tensor, Iterable[int]],
                                        panoptic_divisor: Union[tf.Tensor, int] = 256):
    """Determines which instances to keep after ignoring a set of categories.

    Args:
      panoptic_labels: An integer tensor of panoptic labels of shape `[height,
        width]`. Each element will be `category * panoptic_divisor + instance` for
        a pixel.
      ignore_categories: An iterable or tensor of integer category labels.
        Instances where the category portion of the label in `panoptic_labels` are
        in the ignore set will not be included in the results.
      panoptic_divisor: The divisor used to multiply the category label when
        constructing panoptic labels, as in integer or integer scalar tensor.

    Returns:
      A boolean tensor masking which of the input `panoptic_labels` corresponds
      to an instance that will be kept, or equivalently *not* ignored.
    """
    ignore_categories = tf.convert_to_tensor(
        ignore_categories, dtype=panoptic_labels.dtype)
    panoptic_divisor = tf.convert_to_tensor(
        panoptic_divisor, dtype=panoptic_labels.dtype)

    instance_category = tf.math.floordiv(panoptic_labels, panoptic_divisor)
    instance_is_ignored = tf.math.reduce_any(
        tf.equal(
            tf.expand_dims(instance_category, 1),
            tf.expand_dims(ignore_categories, 0)),
        axis=1)
    instance_is_kept = tf.math.logical_not(instance_is_ignored)

    return instance_is_kept


def _broadcast_over_instances(t, num_instances):
    out_shape = tf.concat([tf.shape(t), [num_instances]], axis=0)
    return tf.broadcast_to(tf.expand_dims(t, -1), out_shape)


def instance_boxes_from_masks(
        panoptic_labels: tf.Tensor,
        ignore_categories: Optional[Union[tf.Tensor, Iterable[int]]] = None,
        panoptic_divisor: Union[tf.Tensor, int] = 256):
    """Finds the bounding boxes around instances, given a panoptic label map.

    Args:
      panoptic_labels: An integer tensor of panoptic labelsof shape `[height,
        width]`. Each element will be `category * panoptic_divisor + instance` for
        a pixel.
      ignore_categories: An iterable or tensor of integer category labels.
        Instances where the category portion of the label in `panoptic_labels` are
        in the ignore set will not be included in the results.
      panoptic_divisor: The divisor used to multiply the category label when
        constructing panoptic labels, as in integer or integer scalar tensor.

    Returns:
      A tuple of arrays (unique_labels, box_coords).
      unique_labels: An tensor of each possible non-ignored label value in
        `panoptic_labels`, in the same order as the boxes.
      box_coords: An tensor of shape `[num_labels, 4]`. Each row is one box as
        `[ymin, xmin, ymax, xmax]`.
    """
    label_shape = tf.shape(panoptic_labels)
    height = label_shape[0]
    width = label_shape[1]
    x_coord, y_coord = tf.meshgrid(
        tf.range(width, dtype=tf.float32), tf.range(height, dtype=tf.float32))

    unique_labels, flat_instance_index = tf.unique(
        tf.reshape(panoptic_labels, [height * width]))
    num_instances = tf.size(unique_labels)
    instance_index = tf.reshape(flat_instance_index, [height, width])

    y_coord_repeated = _broadcast_over_instances(y_coord, num_instances)
    x_coord_repeated = _broadcast_over_instances(x_coord, num_instances)
    instance_index_repeated = _broadcast_over_instances(instance_index,
                                                        num_instances)

    instance_index_matches = tf.math.equal(
        instance_index_repeated,
        tf.reshape(tf.range(num_instances), [1, 1, num_instances]))

    # In these tensors, each slice in the 3rd dimension corresponds to an
    # instance. We replace the pixels that do _not_ belong to that instance with
    # a +/- infinity in order that it not be included in the reduce_min/max below.
    inf3d = tf.broadcast_to([[[float('Inf')]]], tf.shape(x_coord_repeated))
    y_or_inf = tf.where(instance_index_matches, y_coord_repeated, inf3d)
    y_or_neg_inf = tf.where(instance_index_matches, y_coord_repeated, -inf3d)
    x_or_inf = tf.where(instance_index_matches, x_coord_repeated, inf3d)
    x_or_neg_inf = tf.where(instance_index_matches, x_coord_repeated, -inf3d)

    y_min = tf.reduce_min(y_or_inf, axis=[0, 1])
    x_min = tf.reduce_min(x_or_inf, axis=[0, 1])
    y_max = tf.reduce_max(y_or_neg_inf, axis=[0, 1]) + 1
    x_max = tf.reduce_max(x_or_neg_inf, axis=[0, 1]) + 1

    box_coords = tf.stack([y_min, x_min, y_max, x_max], axis=1)

    if ignore_categories is not None:
        # Filter out the boxes that correspond to instances in the "ignore"
        # categories.
        instance_is_kept = instances_without_ignore_categories(
            unique_labels, ignore_categories, panoptic_divisor)

        unique_labels = tf.boolean_mask(unique_labels, instance_is_kept)
        box_coords = tf.boolean_mask(box_coords, instance_is_kept)

    return unique_labels, box_coords


def per_instance_masks(panoptic_labels: tf.Tensor,
                       instance_panoptic_labels: tf.Tensor,
                       out_dtype: tf.dtypes.DType = tf.bool) -> tf.Tensor:
    """3D tensor where each slice in 3rd dimensions is an instance's mask."""
    num_instances = tf.size(instance_panoptic_labels)
    matches = tf.equal(
        tf.expand_dims(panoptic_labels, 0),
        tf.reshape(instance_panoptic_labels, [num_instances, 1, 1]))
    return tf.cast(matches, out_dtype)


def _average_per_instance(map_tensor: tf.Tensor, panoptic_labels: tf.Tensor,
                          instance_panoptic_labels: tf.Tensor,
                          instance_area: tf.Tensor) -> tf.Tensor:
    """Finds the average of the values in map_tensor over each instance."""

    # For each instance (in the 3rd dim), generate a map that has, per-pixel:
    # - The input value if that pixel belongs to the instance.
    # - Zero otherwise.
    pixel_in_instance = per_instance_masks(panoptic_labels,
                                           instance_panoptic_labels)

    map_dtype = map_tensor.dtype
    num_instances = tf.size(instance_panoptic_labels)
    map_or_zero = tf.where(pixel_in_instance, tf.expand_dims(map_tensor, 0),
                           tf.zeros([num_instances, 1, 1], dtype=map_dtype))

    # Average the semantic probabilities over each instance.
    instance_total_prob = tf.math.reduce_sum(map_or_zero, axis=[1, 2])
    instance_avg_prob = tf.divide(instance_total_prob,
                                  tf.cast(instance_area, map_dtype))

    return instance_avg_prob


# pyformat: disable
def per_instance_semantic_probabilities(
        panoptic_labels: tf.Tensor,
        instance_panoptic_labels: tf.Tensor,
        instance_area: tf.Tensor,
        semantic_probability: tf.Tensor,
        panoptic_divisor: Union[tf.Tensor, int],
        ignore_label: Union[tf.Tensor, int]) -> tf.Tensor:
    """Mean probability for the semantic label of each unique instance."""
    # pyformat: enable
    # Get the probability associated with the semantic label encoded in the
    # panoptic_labels at each pixel.
    panoptic_divisor = tf.convert_to_tensor(panoptic_divisor, dtype=tf.int32)
    ignore_label = tf.convert_to_tensor(ignore_label, dtype=tf.int32)
    semantic_label_map = tf.math.floordiv(panoptic_labels, panoptic_divisor)

    map_shape = tf.shape(semantic_label_map)
    height = map_shape[0]
    width = map_shape[1]
    num_pixels = height * width

    semantic_index = tf.reshape(semantic_label_map, [num_pixels])
    # Use 0 as the index for a pixel with the "ignore" label, since that semantic
    # label may not be a valid index into the class axis of the
    # semantic_probability tensor.
    semantic_index = tf.where(semantic_index == ignore_label, 0, semantic_index)

    x, y = tf.meshgrid(tf.range(width), tf.range(height))
    probability_index = tf.stack([
        tf.reshape(y, [num_pixels]),
        tf.reshape(x, [num_pixels]),
        semantic_index,
    ],
        axis=1)

    # When len(semantic_probability.shape) == 3 (e.g., for Panoptic-DeepLab), it
    # is a semantic probability map with shape [H, W, num_classes], so we need to
    # gather the class confidence for corresponding semantic class prediction.
    # When len(semantic_probability.shape) == 2 (e.g., for MaX-DeepLab), it is a
    # semantic confidence map with shape [H, W], so we can directly use it as
    # pixel_semantic_probability.
    if len(semantic_probability.shape) == 3:
        pixel_semantic_probability = tf.reshape(
            tf.gather_nd(semantic_probability, probability_index), [height, width])
    elif len(semantic_probability.shape) == 2:
        pixel_semantic_probability = semantic_probability
    # Set the probability for the "ignore" pixels to 0.
    pixel_semantic_probability = tf.where(semantic_label_map == ignore_label, 0.0,
                                          pixel_semantic_probability)

    instance_avg_prob = _average_per_instance(pixel_semantic_probability,
                                              panoptic_labels,
                                              instance_panoptic_labels,
                                              instance_area)

    return instance_avg_prob


def combined_instance_scores(
        panoptic_labels: tf.Tensor, semantic_probability: tf.Tensor,
        instance_score_map: tf.Tensor, panoptic_divisor: Union[tf.Tensor, int],
        ignore_label: Union[tf.Tensor, int]) -> Tuple[tf.Tensor, tf.Tensor]:
    """Combines (with a product) predicted semantic and instance probabilities.

    Args:
      panoptic_labels: A 2D integer tensor of panoptic format labels (each pixel
        entry is `semantic_label * panoptic_divisor + instance_label`).
      semantic_probability: A 3D float tensor, where the 3rd dimension is over
        semantic labels, and each spatial location will have the discrete
        distribution of the probabilities of the semantic classes.
      instance_score_map: A 2D float tensor, where the pixels for an instance will
        have the probability of that being an instance.
      panoptic_divisor: Integer scalar divisor/multiplier used to construct the
        panoptic labels.
      ignore_label: Integer scalar, for the "ignore" semantic label in the
        panoptic labels.

    Returns:
      A tuple of instance labels and the combined scores for those instances, each
      as a 1D tensor.
    """
    panoptic_divisor = tf.convert_to_tensor(panoptic_divisor, dtype=tf.int32)
    ignore_label = tf.convert_to_tensor(ignore_label, dtype=tf.int32)

    num_pixels = tf.size(panoptic_labels)
    instance_panoptic_labels, _, instance_area = tf.unique_with_counts(
        tf.reshape(panoptic_labels, [num_pixels]))

    instance_semantic_labels = tf.math.floordiv(instance_panoptic_labels,
                                                panoptic_divisor)
    valid_mask = tf.not_equal(instance_semantic_labels, ignore_label)
    instance_panoptic_labels = tf.boolean_mask(instance_panoptic_labels,
                                               valid_mask)
    instance_area = tf.boolean_mask(instance_area, valid_mask)

    instance_semantic_probabilities = per_instance_semantic_probabilities(
        panoptic_labels, instance_panoptic_labels, instance_area,
        semantic_probability, panoptic_divisor, ignore_label)

    instance_scores = _average_per_instance(instance_score_map, panoptic_labels,
                                            instance_panoptic_labels,
                                            instance_area)

    combined_scores = instance_semantic_probabilities * instance_scores
    return instance_panoptic_labels, combined_scores


def per_instance_is_crowd(is_crowd_map: tf.Tensor, id_map: tf.Tensor,
                          output_ids: tf.Tensor) -> tf.Tensor:
    """Determines the per-instance is_crowd value from a boolian is_crowd map.

    Args:
      is_crowd_map: A 2D boolean tensor. Where it is True, the instance in that
        region is a "crowd" instance. It is assumed that all pixels in an instance
        will have the same value in this map.
      id_map: A 2D integer tensor, with the instance id label at each pixel.
      output_ids: A 1D integer vector tensor, the per-instance ids for which to
        output the is_crowd values.

    Returns:
      A 1D boolean vector tensor, with the per-instance is_crowd value. The ith
      element of the return value will be the is_crowd result for the segment
      with the ith element of the output_ids argument.
    """
    flat_is_crowd_map = tf.reshape(is_crowd_map, [-1])
    flat_id_map = tf.reshape(id_map, [-1])

    # Get an is_crowd value from the map for each id.
    # Only need an arbtitrary value due to assumption that the is_crowd map does
    # not vary over an instance.
    unique_ids, unique_index = tf.unique(flat_id_map)
    unique_is_crowd = tf.scatter_nd(
        tf.expand_dims(unique_index, 1), flat_is_crowd_map, tf.shape(unique_ids))

    # Map from the order/set in unique_ids to that in output_ids
    matching_ids = tf.math.equal(
        tf.expand_dims(output_ids, 1), tf.expand_dims(unique_ids, 0))
    matching_index = tf.where(matching_ids)[:, 1]
    return tf.gather(unique_is_crowd, matching_index)



from io import BytesIO
import math
import numpy as np
from PIL import Image
import skimage.draw

from . import python_utils

CV2 = False
if python_utils.module_exists("cv2"):
    import cv2
    CV2 = True


def get_image_size(filepath):
    im = Image.open(filepath)
    return im.size


def load_image(image_filepath):
    image = Image.open(image_filepath)
    image.load()
    image_array = np.array(image, dtype=np.uint8)
    image.close()
    return image_array


def padded_boundingbox(boundingbox, padding):
    boundingbox_new = np.empty_like(boundingbox)
    boundingbox_new[0:2] = boundingbox[0:2] + padding
    boundingbox_new[2:4] = boundingbox[2:4] - padding
    return boundingbox_new


def center_bbox(spatial_shape, output_shape):
    """
    Return a bbox centered in spatial_shape with size output_shape

    :param spatial_shape:
    :param output_shape:
    :return:
    """
    center = (spatial_shape[0] / 2, spatial_shape[1] / 2)
    half_output_shape = (output_shape[0] / 2, output_shape[1] / 2)
    bbox = [
        center[0] -
        half_output_shape[0],
        center[1] -
        half_output_shape[1],
        center[0] +
        half_output_shape[0],
        center[1] +
        half_output_shape[1]]
    bbox = bbox_to_int(bbox)
    return bbox


def bbox_add_margin(bbox, margin):
    bbox_new = bbox.copy()
    bbox_new[0:2] -= margin
    bbox_new[2:4] += margin
    return bbox_new


def bbox_to_int(bbox):
    bbox_new = [
        int(np.floor(bbox[0])),
        int(np.floor(bbox[1])),
        int(np.ceil(bbox[2])),
        int(np.ceil(bbox[3])),
    ]
    return bbox_new


def draw_line_aa_in_patch(edge, patch_bounds):
    rr, cc, prob = skimage.draw.line_aa(
        edge[0][0], edge[0][1], edge[1][0], edge[1][1])
    keep_mask = (patch_bounds[0] <= rr) & (rr < patch_bounds[2]) \
        & (patch_bounds[1] <= cc) & (cc < patch_bounds[3])
    rr = rr[keep_mask]
    cc = cc[keep_mask]
    prob = prob[keep_mask]
    return rr, cc, prob


def convert_array_to_jpg_bytes(image_array, mode=None):
    img = Image.fromarray(image_array, mode=mode)
    output = BytesIO()
    img.save(output, format="JPEG", quality=90)
    contents = output.getvalue()
    output.close()
    return contents


def displacement_map_to_transformation_maps(disp_field_map):
    disp_field_map = disp_field_map.astype(np.float32)
    i = np.arange(disp_field_map.shape[0], dtype=np.float32)
    j = np.arange(disp_field_map.shape[1], dtype=np.float32)
    iv, jv = np.meshgrid(i, j, indexing="ij")
    reverse_map_i = iv + disp_field_map[:, :, 1]
    reverse_map_j = jv + disp_field_map[:, :, 0]
    return reverse_map_i, reverse_map_j


if CV2:
    def apply_displacement_field_to_image(image, disp_field_map):
        trans_map_i, trans_map_j = displacement_map_to_transformation_maps(
            disp_field_map)
        misaligned_image = cv2.remap(
            image, trans_map_j, trans_map_i, cv2.INTER_CUBIC)
        return misaligned_image

    def apply_displacement_fields_to_image(image, disp_field_maps):
        disp_field_map_count = disp_field_maps.shape[0]
        misaligned_image_list = []
        for i in range(disp_field_map_count):
            misaligned_image = apply_displacement_field_to_image(
                image, disp_field_maps[i, :, :, :])
            misaligned_image_list.append(misaligned_image)
        return misaligned_image_list
else:
    def apply_displacement_field_to_image(image, disp_field_map):
        print("cv2 is not available, the apply_displacement_fields_to_image(image, disp_field_map) function cannot work!")

    def apply_displacement_fields_to_image(image, disp_field_maps):
        print("cv2 is not available, the apply_displacement_fields_to_image(image, disp_field_maps) function cannot work!")


def get_axis_patch_count(length, stride, patch_res):
    total_double_padding = patch_res - stride
    patch_count = max(
        1, int(
            math.ceil(
                (length - total_double_padding) / stride)))
    return patch_count


def compute_patch_boundingboxes(image_size, stride, patch_res):
    """

    @param image_size:
    @param stride:
    @param patch_res:
    @return: [[row_start, col_start, row_end, col_end], ...]
    """
    im_rows = image_size[0]
    im_cols = image_size[1]

    row_patch_count = get_axis_patch_count(im_rows, stride, patch_res)
    col_patch_count = get_axis_patch_count(im_cols, stride, patch_res)

    patch_boundingboxes = []
    for i in range(0, row_patch_count):
        if i < row_patch_count - 1:
            row_slice_begin = i * stride
            row_slice_end = row_slice_begin + patch_res
        else:
            row_slice_end = im_rows
            row_slice_begin = row_slice_end - patch_res
        for j in range(0, col_patch_count):
            if j < col_patch_count - 1:
                col_slice_begin = j * stride
                col_slice_end = col_slice_begin + patch_res
            else:
                col_slice_end = im_cols
                col_slice_begin = col_slice_end - patch_res

            patch_boundingbox = np.array(
                [row_slice_begin, col_slice_begin, row_slice_end, col_slice_end], dtype=np.int)
            assert row_slice_end - row_slice_begin == col_slice_end - \
                col_slice_begin == patch_res, "ERROR: patch does not have the requested shape"
            patch_boundingboxes.append(patch_boundingbox)

    return patch_boundingboxes


def clip_boundingbox(boundingbox, clip_list):
    assert len(boundingbox) == len(
        clip_list), "len(boundingbox) should be equal to len(clip_values)"
    clipped_boundingbox = []
    for bb_value, clip in zip(boundingbox[:2], clip_list[:2]):
        clipped_value = max(clip, bb_value)
        clipped_boundingbox.append(clipped_value)
    for bb_value, clip in zip(boundingbox[2:], clip_list[2:]):
        clipped_value = min(clip, bb_value)
        clipped_boundingbox.append(clipped_value)
    return clipped_boundingbox


def crop_or_pad_image_with_boundingbox(image, patch_boundingbox):
    im_rows = image.shape[0]
    im_cols = image.shape[1]

    row_padding_before = max(0, - patch_boundingbox[0])
    col_padding_before = max(0, - patch_boundingbox[1])
    row_padding_after = max(0, patch_boundingbox[2] - im_rows)
    col_padding_after = max(0, patch_boundingbox[3] - im_cols)

    # Center padding:
    row_padding = row_padding_before + row_padding_after
    col_padding = col_padding_before + col_padding_after
    row_padding_before = row_padding // 2
    col_padding_before = col_padding // 2
    row_padding_after = row_padding - row_padding // 2
    col_padding_after = col_padding - col_padding // 2

    clipped_patch_boundingbox = clip_boundingbox(
        patch_boundingbox, [0, 0, im_rows, im_cols])

    if len(image.shape) == 2:
        patch = image[clipped_patch_boundingbox[0]:clipped_patch_boundingbox[2],
                      clipped_patch_boundingbox[1]:clipped_patch_boundingbox[3]]
        patch = np.pad(patch,
                       [(row_padding_before,
                         row_padding_after),
                        (col_padding_before,
                         col_padding_after)],
                       mode="constant")
    elif len(image.shape) == 3:
        patch = image[clipped_patch_boundingbox[0]:clipped_patch_boundingbox[2],
                      clipped_patch_boundingbox[1]:clipped_patch_boundingbox[3], :]
        patch = np.pad(patch,
                       [(row_padding_before,
                         row_padding_after),
                        (col_padding_before,
                         col_padding_after),
                           (0,
                            0)],
                       mode="constant")
    else:
        print("Image input does not have the right shape/")
        patch = None
    return patch


def make_grid(images, padding=2, pad_value=0, return_offsets=False):
    nmaps = images.shape[0]
    ymaps = int(math.floor(math.sqrt(nmaps)))
    xmaps = nmaps // ymaps
    height, width = int(
        images.shape[1] + padding), int(images.shape[2] + padding)
    grid = np.zeros(
        (height * ymaps + padding,
         width * xmaps + padding,
         images.shape[3])) + pad_value
    k = 0
    offsets = []
    for y in range(ymaps):
        for x in range(xmaps):
            if k >= nmaps:
                break
            x_offset = x * width + padding
            y_offset = y * height + padding
            grid[y * height + padding:(y + 1) * height,
                 x * width + padding:(x + 1) * width,
                 :] = images[k]
            offsets.append((x_offset, y_offset))
            k = k + 1
    if return_offsets:
        return grid, offsets
    else:
        return grid


def main():
    if python_utils.module_exists("matplotlib.pyplot"):
        import matplotlib.pyplot as plt
    im_rows = 5
    im_cols = 10
    stride = 1
    patch_res = 15

    image = np.random.randint(0, 256, size=(
        im_rows, im_cols, 3), dtype=np.uint8)
    image = Image.fromarray(image)
    image = np.array(image)
    plt.ion()
    plt.figure(1)
    plt.imshow(image)
    plt.show()

    # Cut patches
    patch_boundingboxes = compute_patch_boundingboxes(
        image.shape[0:2], stride, patch_res)

    plt.figure(2)

    for patch_boundingbox in patch_boundingboxes:
        patch = crop_or_pad_image_with_boundingbox(image, patch_boundingbox)
        plt.imshow(patch)
        plt.show()
        input("Press <Enter> to finish...")


if __name__ == "__main__":
    main()



from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import SPACETIME_CACHE_VERSION  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import SpacetimeGridBuilder  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _build_grid  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import ImagePropertyCacher  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _sample_single_video_spacetime_targets  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _build_targets_around_track  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _build_targets_in_spatial_region  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _build_vidspace_track_qtree  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _refine_time_sample  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import _visualize_sample_grid  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.spacetime_grid_builder import sample_video_spacetime_targets  # NOQA



"""
This module was moved to kwgis.utils.util_gdal, prefer using that directly.

SeeAlso
    util_raster.py

References:
    https://gdal.org/programs/gdalwarp.html#cmdoption-gdalwarp-multi
    https://gis.stackexchange.com/a/241810
    https://trac.osgeo.org/gdal/wiki/UserDocs/GdalWarp#WillincreasingRAMincreasethespeedofgdalwarp
    https://github.com/OpenDroneMap/ODM/issues/778


TODO:
    TODO test this and see if it's safe to add:
        --config GDAL_PAM_ENABLED NO
    Removes .aux.xml sidecar files and puts them in the geotiff metadata
    ex. histogram from fmask
    https://stackoverflow.com/a/51075774
    https://trac.osgeo.org/gdal/wiki/ConfigOptions#GDAL_PAM_ENABLED
    https://gdal.org/drivers/raster/gtiff.html#georeferencing
"""
import ubelt as ub


def __getattr__(key):
    ub.schedule_deprecation(
        'geowatch', key, 'attribute of util_gdal',
        migration='use kwgis.util_gdal directly instead',
        deprecate='0.18.3', error='0.20.0', remove='0.21.0'
    )
    from kwgis.utils import util_gdal
    return getattr(util_gdal, key)



"""
Quick and dirty project specific stuff that ideally wont get in the way of
general use-cases but should eventually be factored out.

Special heuristics. Used by ./aggregate.py and ./aggregate_plots.py
"""
import kwarray
import ubelt as ub
from kwutil.util_yaml import Yaml


class SmartGlobalHelper:
    """
    A class for SMART-specific hacks and defaults for mlops

    Should be stateless
    """

    def __init__(self):
        import kwimage
        self.delivery_to_color = {
            # 'Eval7': kwimage.Color('kitware_yellow').as01(),
            # 'Eval8': kwimage.Color('kitware_orange').as01(),
            'Eval9': kwimage.Color('purple').as01(),
            'Eval10': kwimage.Color('kitware_darkblue').as01(),
            'Eval11': kwimage.Color('kitware_orange').as01(),
            'Eval13': kwimage.Color('kitware_blue').as01(),
            'Baseline2023-07': kwimage.Color('kitware_green').as01(),
        }

    # BAA References
    # https://sam.gov/opp/e56a9469354f4ffb9756b59a174fcdc2/view
    TARGET_METRICS = {
        'bas_tpr': 0.59,
        'bas_ppv': 0.85,
        'bas_fpr': 0.1,
        'bas_f1': 0.7,
        'cost_dollars_per_km2': 30,
        'ac_time_delta': "5 days",
    }

    VIZ_BLOCKLIST = {
        'resolved_params.bas_poly_eval.pred_sites',
        'resolved_params.bas_poly_eval.gt_dpath',
        'resolved_params.bas_poly_eval.true_site_dpath',
        'resolved_params.bas_poly_eval.true_region_dpath',
        'resolved_params.bas_poly_eval.out_dir',
        'resolved_params.bas_poly_eval.merge',
        'resolved_params.bas_poly_eval.merge_fpath',
        'resolved_params.bas_poly_eval.merge_fbetas',
        'resolved_params.bas_poly_eval.tmp_dir',
        'resolved_params.bas_poly_eval.enable_viz',
        'resolved_params.bas_poly_eval.name',
        'resolved_params.bas_poly_eval.use_cache',
        'resolved_params.bas_poly_eval.load_workers',
        'resolved_params.bas_poly.in_file',
        'resolved_params.bas_poly.out_kwcoco',
        'resolved_params.bas_poly.out_sites_dir',
        'resolved_params.bas_poly.out_site_summaries_dir',
        'resolved_params.bas_poly.out_sites_fpath',
        'resolved_params.bas_poly.out_site_summaries_fpath',
        'resolved_params.bas_poly.in_file_gt',
        'resolved_params.bas_poly.region_id',
        'resolved_params.bas_poly.default_track_fn',
        'resolved_params.bas_poly.site_summary',
        'resolved_params.bas_poly.clear_annots',
        'resolved_params.bas_poly.append_mode',
        'resolved_params.bas_pxl.config_file',
        'resolved_params.bas_pxl.write_out_config_file_to_this_path',
        'resolved_params.bas_pxl.datamodule',
        'resolved_params.bas_pxl.pred_dataset',
        'resolved_params.bas_pxl.devices',
        'resolved_params.bas_pxl.with_change',
        'resolved_params.bas_pxl.with_class',
        'resolved_params.bas_pxl.with_saliency',
        'resolved_params.bas_pxl.compress',
        'resolved_params.bas_pxl.track_emissions',
        'resolved_params.bas_pxl.quantize',
        'resolved_params.bas_pxl.clear_annots',
        'resolved_params.bas_pxl.write_workers',
        'resolved_params.bas_pxl.write_preds',
        'resolved_params.bas_pxl.write_probs',
        'resolved_params.bas_pxl.train_dataset',
        'resolved_params.bas_pxl.vali_dataset',
        'resolved_params.bas_pxl.test_dataset',
        'resolved_params.bas_pxl.batch_size',
        'resolved_params.bas_pxl.normalize_inputs',
        'resolved_params.bas_pxl.num_workers',
        'resolved_params.bas_pxl.torch_sharing_strategy',
        'resolved_params.bas_pxl.torch_start_method',
        'resolved_params.bas_pxl.sqlview',
        'resolved_params.bas_pxl.max_epoch_length',
        'resolved_params.bas_pxl.use_centered_positives',
        'resolved_params.bas_pxl.use_grid_positives',
        'resolved_params.bas_pxl.use_grid_valid_regions',
        'resolved_params.bas_pxl.neg_to_pos_ratio',
        'resolved_params.bas_pxl.use_grid_cache',
        'resolved_params.bas_pxl.ignore_dilate',
        'resolved_params.bas_pxl.weight_dilate',
        'resolved_params.bas_pxl.min_spacetime_weight',
        'resolved_params.bas_pxl.upweight_centers',
        'resolved_params.bas_pxl.upweight_time',
        'resolved_params.bas_pxl.dist_weights',
        'resolved_params.bas_pxl.balance_areas',
        'resolved_params.bas_pxl.resample_invalid_frames',
        'resolved_params.bas_pxl.downweight_nan_regions',
        'resolved_params.bas_pxl.temporal_dropout',
        'resolved_params.bas_pxl_fit.accelerator',
        'resolved_params.bas_pxl_fit.accumulate_grad_batches',
        'resolved_params.bas_pxl_fit.datamodule',
        'resolved_params.bas_pxl_fit.devices',
        'resolved_params.bas_pxl_fit.gradient_clip_algorithm',
        'resolved_params.bas_pxl_fit.gradient_clip_val',
        'resolved_params.bas_pxl_fit.max_epochs',
        'resolved_params.bas_pxl_fit.max_steps',
        'resolved_params.bas_pxl_fit.method',
        'resolved_params.bas_pxl_fit.name',
        'resolved_params.bas_pxl_fit.patience',
        'resolved_params.bas_pxl_fit.precision',
        'resolved_params.bas_pxl_fit.sqlview',
        'resolved_params.bas_pxl_fit.stochastic_weight_avg',
        'resolved_params.bas_pxl_fit.inference_mode',
        'resolved_params.bas_pxl_fit.use_grid_cache',
        'resolved_params.bas_pxl_fit.use_grid_valid_regions',
        'resolved_params.bas_pxl_eval.balance_area',
        'resolved_params.bas_pxl_eval.draw_curves',
        'resolved_params.bas_pxl_eval.draw_heatmaps',
        'resolved_params.bas_pxl_eval.draw_workers',
        'resolved_params.bas_pxl_eval.eval_dpath',
        'resolved_params.bas_pxl_eval.eval_fpath',
        'resolved_params.bas_pxl_eval.pred_dataset',
        'resolved_params.bas_pxl_eval.resolution',
        'resolved_params.bas_pxl_eval.score_space',
        'resolved_params.bas_pxl_eval.true_dataset',
        'resolved_params.bas_pxl_eval.viz_thresh',
        'resolved_params.bas_pxl_eval.workers',
    }

    EXTRA_HASHID_IGNORE_COLUMNS = [
        'params.sc_poly.site_summary',
        'params.sc_pxl.num_workers',
        'params.bas_pxl.num_workers',
    ]

    # Mark columns that are typically paths. Used when building effective params.
    EXTRA_PATH_COLUMNS = [
        'params.bas_poly_eval.true_site_dpath',
        'params.bas_poly_eval.true_region_dpath',
        'params.bas_poly.boundary_region',
    ]

    LABEL_MAPPINGS = {
        'region_id': 'Region',
        'metrics.sc_poly_eval.bas_ffpa': 'FFPA',
        'metrics.sc_poly_eval.bas_faa_f1': 'BAS-FAA-F1',
        'sc_poly_eval.bas_ffpa': 'FFPA',
        'sc_poly_eval.bas_faa_f1': 'BAS-FAA-F1',

        'metrics.sc_poly_eval.bas_f1': 'BAS-F1',
        'metrics.sc_poly_eval.sc_macro_f1': 'AC-F1 (macro)',
        'sc_poly_eval.bas_f1': 'BAS-F1',
        'sc_poly_eval.sc_macro_f1': 'AC-F1 (macro)',
    }

    def shared_palettes(self, macro_table):
        """
        For each key in a hard code set (relevant to SMART), assign a
        consistent color to those values so our plots are comparble.
        """
        # import kwplot
        import numpy as np
        import seaborn as sns
        from geowatch.utils import util_pandas

        def filterkeys(keys):
            return [k for k in keys if 'specified' not in k]

        _macro_table = util_pandas.DotDictDataFrame(macro_table)
        thresh_keys = filterkeys(_macro_table.query_column('thresh'))
        bas_thresh_keys = [k for k in thresh_keys if 'bas_poly' in k]

        # Pre determine some palettes
        shared_palletes_groups = []
        shared_palletes_groups.append(bas_thresh_keys)

        shared_palletes_groups.append(filterkeys(_macro_table.query_column('learning_rate')))

        shared_palletes_groups.append(filterkeys(_macro_table.query_column('chip_dims')))
        shared_palletes_groups.append(
            filterkeys(_macro_table.query_column('resolution')) +
            filterkeys(_macro_table.query_column('output_space_scale')) +
            filterkeys(_macro_table.query_column('output_resolution')) +
            filterkeys(_macro_table.query_column('input_space_scale')) +
            filterkeys(_macro_table.query_column('input_resolution')) +
            filterkeys(_macro_table.query_column('window_space_scale')) +
            filterkeys(_macro_table.query_column('window_resolution'))
        )

        param_to_palette = {}
        for group_params in shared_palletes_groups:
            try:
                unique_vals = np.unique(macro_table[group_params].values)
            except TypeError:
                unique_vals = set(macro_table[group_params].values.ravel())

            if len(unique_vals) > 5:
                unique_colors = sns.color_palette('Spectral', n_colors=len(unique_vals))
                # kwplot.imshow(_draw_color_swatch(unique_colors), fnum=32)
            else:
                unique_colors = sns.color_palette(n_colors=len(unique_vals))
            palette = ub.dzip(unique_vals, unique_colors)
            param_to_palette.update({p: palette for p in group_params})
        return param_to_palette

    def make_param_palette(self, param_values):
        import numpy as np
        import seaborn as sns
        try:
            unique_vals = np.unique(param_values)
        except TypeError:
            unique_vals = set(param_values.ravel())

        # 'Spectral'
        if False and len(unique_vals) > 5:
            unique_colors = sns.color_palette('Spectral', n_colors=len(unique_vals))
            # kwplot.imshow(_draw_color_swatch(unique_colors), fnum=32)
        else:
            unique_colors = sns.color_palette(n_colors=len(unique_vals))
        palette = ub.dzip(unique_vals, unique_colors)
        return palette

    def region_palette(self, rois=None):
        if rois is None:
            rois = [
                'AE_R001',
                'KR_R001',
                'NZ_R001',
                'CH_R001',
                'BR_R002',

                'KR_R002',

                'CN_C000',
                'CO_C001',
                'KW_C001',
                'SA_C001',
                'VN_C002',
            ]
        from geowatch.utils import util_kwplot
        roi_to_color = util_kwplot.Palette.coerce({
            'KR_R002': (230,  76, 230),
        })
        # plotter = aggregate_plots.build_plotter(agg, rois, plot_config)
        roi_to_color.update(rois)
        roi_to_color = roi_to_color.reorder(rois)
        return roi_to_color

    def label_modifier(self):
        """
        Build the label modifier for the SMART task.

        Returns:
            util_kwplot.LabelModifier
        """
        from geowatch.utils import util_kwplot
        modifier = util_kwplot.LabelModifier()

        modifier.add_mapping({
            'blue|green|red|nir': 'BGRN',
            'blue|green|red|nir,invariants.0:17': 'invar',
            'blue|green|red|nir|swir16|swir22': 'BGNRSH'
        })
        modifier.add_mapping(self._humanize_label)
        return modifier

    @staticmethod
    def _humanize_label(text):
        text = text.replace('package_epoch0_step41', 'EVAL7')
        text = text.replace('params.', '')
        text = text.replace('metrics.', '')
        text = text.replace('fit.', 'fit.')
        return text

    def default_vantage_points(self, eval_type):
        if eval_type == 'bas_poly_eval':
            vantage_points = [
                {
                    'metric1': 'metrics.bas_poly_eval.bas_faa_f1',
                    'metric2': 'metrics.bas_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },
                # {
                #     'metric1': 'metrics.bas_pxl_eval.salient_AP',
                #     'metric2': 'metrics.bas_poly_eval.bas_tpr',

                #     'scale1': 'linear',
                #     'scale2': 'linear',

                #     'objective1': 'maximize',
                # },

                {
                    'metric1': 'metrics.bas_poly_eval.bas_tpr',
                    'metric2': 'metrics.bas_poly_eval.bas_f1',

                    'objective1': 'maximize',

                    'scale1': 'linear',
                    'scale2': 'linear',
                },

                {
                    'metric1': 'metrics.bas_poly_eval.bas_ppv',
                    'metric2': 'metrics.bas_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },

                {
                    'metric1': 'metrics.bas_poly_eval.bas_f1',
                    'metric2': 'metrics.bas_poly_eval.bas_ffpa',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },

                {
                    'metric1': 'metrics.bas_poly_eval.bas_space_FAR',
                    'metric2': 'metrics.bas_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'minimize',
                },
            ]
        elif eval_type == 'sv_poly_eval':
            vantage_points = [
                {
                    'metric1': 'metrics.sv_poly_eval.bas_faa_f1',
                    'metric2': 'metrics.sv_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },

                {
                    'metric1': 'metrics.sv_poly_eval.bas_tpr',
                    'metric2': 'metrics.sv_poly_eval.bas_f1',

                    'objective1': 'maximize',

                    'scale1': 'linear',
                    'scale2': 'linear',
                },

                {
                    'metric1': 'metrics.sv_poly_eval.bas_ppv',
                    'metric2': 'metrics.sv_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },

                {
                    'metric1': 'metrics.sv_poly_eval.bas_f1',
                    'metric2': 'metrics.sv_poly_eval.bas_ffpa',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },

                {
                    'metric1': 'metrics.sv_poly_eval.bas_space_FAR',
                    'metric2': 'metrics.sv_poly_eval.bas_tpr',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'minimize',
                },
            ]
        elif eval_type == 'bas_pxl_eval':
            vantage_points = [
                {
                    'metric1': 'metrics.bas_pxl_eval.salient_AP',
                    'metric2': 'metrics.bas_pxl_eval.salient_AUC',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },
            ]
        elif eval_type == 'sc_poly_eval':
            vantage_points = [
                {
                    'metric1': 'metrics.sc_poly_eval.sc_macro_f1',
                    'metric2': 'metrics.sc_poly_eval.bas_faa_f1',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },
                {
                    'metric1': 'metrics.sc_poly_eval.bas_f1',
                    'metric2': 'metrics.sc_poly_eval.bas_ffpa',

                    'scale1': 'linear',
                    'scale2': 'linear',

                    'objective1': 'maximize',
                },
            ]

        return vantage_points

    def _default_metrics(self, agg):
        _display_metrics_suffixes = []
        if agg.node_type in { 'bas_poly_eval', 'sv_poly_eval'}:
            _display_metrics_suffixes = [
                'bas_tp',
                'bas_fp',
                'bas_fn',
                'bas_tpr',
                'bas_f1',
                'bas_ffpa',
                'bas_faa_f1',
                # 'bas_tpr',
                # 'bas_ppv',
            ]
            _primary_metrics_suffixes = [
                # 'bas_faa_f1'
                'bas_faa_f1',
            ]
        elif agg.node_type == 'sc_poly_eval':
            _display_metrics_suffixes = [
                'macro_f1_siteprep',
                'macro_f1_active',
                'bas_tp',
                'bas_fp',
                'bas_fn',
                'bas_tpr',
                'bas_f1',
                'bas_ffpa',
                'bas_faa_f1',
            ]
            _primary_metrics_suffixes = [
                'bas_faa_f1', 'sc_macro_f1',
            ]
        elif agg.node_type == 'bas_pxl_eval':
            _primary_metrics_suffixes = [
                'salient_AP',
                # 'salient_APUC',
                'salient_AUC',
            ]
        elif agg.node_type == 'sc_pxl_eval':
            _primary_metrics_suffixes = [
                'coi_mAP',
                # 'coi_mAPUC',
                'coi_mAUC',
            ]
        else:
            raise NotImplementedError(agg.node_type)
        return _primary_metrics_suffixes, _display_metrics_suffixes

    def mark_star_models(self, macro_table):
        #### Hack for models of interest.
        star_params = []
        p1 = macro_table[(
            # (macro_table['bas_poly.moving_window_size'] == 200) &
            (macro_table['params.bas_pxl.package_fpath'] == 'package_epoch0_step41') &
            (macro_table['params.bas_pxl.chip_dims'] == '[128, 128]') &
            (macro_table['params.bas_poly.thresh'] == 0.12)  &
            (macro_table['params.bas_poly.max_area_sqkm'] == 'None') &
            (macro_table['params.bas_poly.moving_window_size'] == 'None')
        )]['param_hashid'].iloc[0]
        star_params = [p1]
        p2 = macro_table[(
            # (macro_table['bas_poly.moving_window_size'] == 200) &
            (macro_table['params.bas_pxl.package_fpath'] == 'Drop4_BAS_2022_12_15GSD_BGRN_V10_epoch=0-step=4305') &
            (macro_table['params.bas_pxl.chip_dims'] == '[224, 224]') &
            (macro_table['params.bas_poly.thresh'] == 0.13)  &
            (macro_table['params.bas_poly.max_area_sqkm'] == 'None') &
            (macro_table['params.bas_poly.moving_window_size'] == 200)
        )]['param_hashid'].iloc[0]
        star_params += [p2]
        p3 = macro_table[(
            # (macro_table['bas_poly.moving_window_size'] == 200) &
            (macro_table['params.bas_pxl.package_fpath'] == 'Drop4_BAS_15GSD_BGRNSH_invar_V8_epoch=16-step=8704') &
            (macro_table['params.bas_pxl.chip_dims'] == '[256, 256]') &
            (macro_table['params.bas_poly.thresh'] == 0.17)  &
            (macro_table['params.bas_poly.max_area_sqkm'] == 'None') &
            (macro_table['params.bas_poly.moving_window_size'] == 'None')
        )]['param_hashid'].iloc[0]
        star_params += [p3]
        macro_table['is_star'] = kwarray.isect_flags(macro_table['param_hashid'], star_params)

    def mark_delivery(self, table, include=None):
        """
        self = SMART_HELPER
        """
        from geowatch.utils.util_pandas import DotDictDataFrame
        delivered_model_params = self.get_delivered_model_params()

        delivered_params = delivered_model_params[-1]
        table['delivery'] = None
        table['delivered_params']  = None
        for delivered_params in delivered_model_params:
            if include is not None:
                if delivered_params['delivery'] not in include:
                    continue
            if delivered_params['task'] == 'BAS':
                try:
                    is_delivered_model = table['resolved_params.bas_pxl.package_fpath'].str.endswith(delivered_params['bas_pxl.package_fpath'])
                except Exception:
                    is_delivered_model = table['resolved_params.bas_pxl.package_fpath'].apply(str).str.endswith(delivered_params['bas_pxl.package_fpath'])
                is_delivered_model = is_delivered_model.fillna(False)

                if is_delivered_model.any():
                    table.loc[is_delivered_model, 'delivery_model'] = delivered_params['delivery']

                if True:
                    subset = table[is_delivered_model]
                    failed_subsets = []
                    keys = [
                        'bas_poly.thresh',
                        'bas_pxl.chip_dims',
                        'bas_poly.moving_window_size',
                        'bas_poly.min_area_square_meters',
                        'bas_poly.norm_ord',
                        'bas_poly.poly_merge_method',
                    ]
                    for key in keys:
                        if key in delivered_params:
                            flags = subset['resolved_params.' + key] == delivered_params[key]
                            if flags.sum() == 0:
                                failed_subsets.append(key)
                                print(f'query {key} failed')
                            else:
                                subset = subset[flags]

                    varied = ub.varied_values(DotDictDataFrame(subset)['resolved_params.bas_poly'].to_dict('records'), min_variations=2)
                    varied.pop('resolved_params.bas_poly.out_sites_fpath', None)
                    varied.pop('resolved_params.bas_poly.out_kwcoco', None)
                    varied.pop('resolved_params.bas_poly.in_file', None)
                    varied.pop('resolved_params.bas_poly.boundary_region', None)
                    varied.pop('resolved_params.bas_poly.out_site_summaries_dir', None)
                    varied.pop('resolved_params.bas_poly.out_sites_dir', None)
                    varied.pop('resolved_params.bas_poly.out_site_summaries_fpath', None)
                    if failed_subsets or varied:
                        print('Failed to get full spec on: ' + ub.urepr(failed_subsets))
                        print('varied = {}'.format(ub.urepr(varied, nl=1)))
                        for key in failed_subsets:
                            print(f'key={key}')
                            print(subset['resolved_params.' + key].unique())

                    table.loc[subset.index, 'delivered_params'] = delivered_params['delivery']

                if 0:
                    ub.varied_values(DotDictDataFrame(subset)['resolved_params.bas_poly'].to_dict('records'), min_variations=2).keys()
                    DotDictDataFrame(subset)['resolved_params.bas_pxl']

            delivered_marked = table['delivered_params'].unique()
            print('delivered_marked = {}'.format(ub.urepr(delivered_marked, nl=1)))

    def populate_test_dataset_bundles(self, agg):
        """
        Attempt to parse out which kwcoco bundle test datasets belonged to
        """
        import os
        test_datasets = agg.table[agg.test_dset_cols]
        dataset_to_bundle = {}
        for key, values in test_datasets.to_dict('list').items():
            for value in ub.unique(values):
                if isinstance(value, (str, os.PathLike)):
                    path = ub.Path(value)
                    bundle_name = path.parent.name
                    dataset_to_bundle[value] = bundle_name

        new_columns = {}
        for key in agg.test_dset_cols:
            new_columns[key] = agg.table[key].apply(lambda x: dataset_to_bundle.get(x, 'unknown'))

        test_bundle_cols = []

        for key, vals in new_columns.items():
            new_key = key + '_bundle'
            new_key_suffix = new_key.split('.', 1)[1]
            new_key2 = 'specified.params.' + new_key_suffix
            new_key3 = 'resolved_params.' + new_key_suffix
            test_bundle_cols.append(new_key)
            agg.table[new_key] = vals
            agg.table[new_key2] = False
            agg.table[new_key3] = vals
        return test_bundle_cols

    def get_delivered_model_params(self):
        delivered_model_params = []
        delivered_model_params += [
            {
                'delivery': 'Eval6',  # ?
                'name': 'Drop4_BAS_Continue_15GSD_BGR_V004_epoch=78-step=323584.pt.pt',
                'bas_pxl.package_fpath': 'models/fusion/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/packages/Drop4_BAS_Continue_15GSD_BGR_V004/Drop4_BAS_Continue_15GSD_BGR_V004_epoch=78-step=323584.pt.pt',
                'task': 'BAS',
            },
            # Phase2 Eval: 2020-11-21
            {
                'delivery': 'Eval7',
                'name': 'package_epoch0_step41.pt.pt',
                'bas_pxl.package_fpath': 'models/fusion/Drop4-BAS/packages/Drop4_TuneV323_BAS_30GSD_BGRNSH_V2/package_epoch0_step41.pt.pt',
                # 'bas_poly.thresh': 0.12,  # hack, I think this was the real one, but we dont have that eval
                'bas_poly.thresh': 0.16,
                'bas_pxl.chip_dims': '[128, 128]',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.min_area_square_meters': 72000.0,
                'task': 'BAS',
            },
            {
                'delivery': 'Eval8',
                'bas_pxl.package_fpath': 'models/fusion/Drop4-BAS/packages/Drop4_BAS_15GSD_BGRNSH_invar_V8/Drop4_BAS_15GSD_BGRNSH_invar_V8_epoch=16-step=8704.pt',
                'task': 'BAS',
                'bas_poly.thresh': 0.17,
                # 'bas_pxl.chip_dims': '[256, 256]',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.min_area_square_meters': 72000.0,
            },

            {
                'delivery': 'Eval9',
                'bas_pxl.package_fpath': 'models/fusion/Drop4-BAS/packages/Drop4_BAS_15GSD_BGRNSH_invar_V8/Drop4_BAS_15GSD_BGRNSH_invar_V8_epoch=16-step=8704.pt',
                'task': 'BAS',
                'bas_poly.thresh': 0.17,
                # 'bas_pxl.chip_dims': '[256, 256]',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.min_area_square_meters': 7200.0,
            },
            ###
            # Eval9
            # bas_tracking_config = {
            #     "thresh": bas_thresh,
            #     "moving_window_size": None,
            #     "polygon_simplify_tolerance": 1,
            #     "min_area_square_meters": 7200,
            #     "resolution": 8,  # Should match "window_space_scale" in SC fusion parameters  # noqa
            #     "max_area_behavior": 'ignore'}
            # {
            #       "chip_overlap": 0.3,
            #       "chip_dims": "auto",
            #       "time_span": "auto",
            #       "time_sampling": "auto",
            #       "drop_unused_frames": true
            # }

            ###


            {
                'delivery': 'Eval10',
                'bas_pxl.package_fpath': 'models/fusion/Drop6-MeanYear10GSD/packages/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2_epoch3_step941.pt',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_poly.thresh': 0.33,
                'bas_poly.moving_window_size': 'None',
                'bas_poly.min_area_square_meters': 7200.0,
                'bas_poly.norm_ord': float('inf'),
                'bas_poly.poly_merge_method': 'v2',
                'task': 'BAS',
            },

            {
                'delivery': 'Eval11',
                'bas_pxl.package_fpath': 'models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_pxl.time_sampling': 'soft4',
                'bas_pxl.tta_fliprot': 3,
                'bas_pxl.tta_time': 3,
                'bas_poly.thresh': 0.425,
                'bas_poly.time_thresh': 0.8,
                'bas_poly.inner_window_size': '1y',
                'bas_poly.inner_agg_fn': 'mean',
                'bas_poly.agg_fn': 'probs',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.polygon_simplify_tolerance': 1,
                'bas_poly.norm_ord': float('inf'),
                'bas_poly.poly_merge_method': 'v2',
                'bas_poly.min_area_square_meters': 7200.0,
                'bas_poly.max_area_square_meters': 8000000.0,
                'task': 'BAS',
            },

            {
                'delivery': 'Eval13',
                'bas_pxl.package_fpath': 'models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_pxl.time_sampling': 'soft4',
                'bas_pxl.tta_fliprot': 1,
                'bas_pxl.tta_time': 1,
                'bas_poly.thresh': 0.40,
                'bas_poly.time_thresh': 0.8,
                'bas_poly.inner_window_size': '1y',
                'bas_poly.inner_agg_fn': 'mean',
                'bas_poly.agg_fn': 'probs',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.polygon_simplify_tolerance': 1,
                'bas_poly.norm_ord': float('inf'),
                'bas_poly.poly_merge_method': 'v2',
                'bas_poly.min_area_square_meters': 7200.0,
                'bas_poly.max_area_square_meters': 8000000.0,
                'task': 'BAS',
            },

            {
                'delivery': 'Baseline2023-07',
                'bas_pxl.package_fpath': 'models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt',
                'bas_pxl.chip_dims': '[196, 196]',
                'bas_pxl.time_sampling': 'soft4',
                'bas_poly.thresh': 0.425,
                'bas_poly.time_thresh': 0.8,
                'bas_poly.inner_window_size': '1y',
                'bas_poly.inner_agg_fn': 'max',
                'bas_poly.agg_fn': 'probs',
                'bas_poly.moving_window_size': 'None',
                'bas_poly.polygon_simplify_tolerance': 1,
                'bas_poly.norm_ord': float('inf'),
                'bas_poly.poly_merge_method': 'v2',
                'bas_poly.min_area_square_meters': 7200.0,
                'bas_poly.max_area_square_meters': 8000000.0,
                'task': 'BAS',
            },
        ]

        yamls_items = [
            (
                '''
                delivery: Eval14
                dag: KIT_TA2_PREEVAL14_BATCH_V27.py
                bas_pxl.package_fpath: models/fusion/Drop7-MedianNoWinter10GSD/packages/Drop7-MedianNoWinter10GSD_bgr_cold_split6_V62/Drop7-MedianNoWinter10GSD_bgr_cold_split6_V62_epoch359_step15480.pt
                sc_pxl.package_fpath: 'models/fusion/Drop4-SC/packages/Drop4_tune_V30_8GSD_V3/Drop4_tune_V30_8GSD_V3_epoch=2-step=17334.pt.pt'

                bas_poly.thresh: 0.3875
                bas_poly.time_thresh: 0.8
                sv_dino_filter.end_min_score: 0.1
                sv_depth_score.model_fpath: models/depth_pcd/basicModel2.h5
                sv_depth_filter.threshold: 0.1

                bas_pxl.chip_overlap: 0.3
                bas_pxl.chip_dims: 196,196
                bas_pxl.time_span: auto
                bas_pxl.fixed_resolution: 10GSD
                bas_pxl.time_sampling: soft4
                bas_pxl.tta_fliprot: 3
                bas_pxl.tta_time: 3
                bas_poly.inner_window_size: 1y
                bas_poly.inner_agg_fn: mean
                bas_poly.norm_ord: inf
                bas_poly.resolution: 10GSD
                bas_poly.moving_window_size: null
                bas_poly.poly_merge_method: v2
                bas_poly.polygon_simplify_tolerance: 1
                bas_poly.agg_fn: probs
                bas_poly.min_area_square_meters: 7200
                bas_poly.max_area_square_meters: 8000000
                sv_dino_filter.box_isect_threshold: 0.1
                sv_dino_filter.box_score_threshold: 0.01
                sv_dino_filter.start_max_score: 1
                '''
            ),
            (
                '''
                delivery: Eval16
                dag: KIT_TA2_PREEVAL16_BATCH_V62.py
                bas_pxl.package_fpath: 'models/fusion/Drop7-MedianNoWinter10GSD/packages/Drop7-MedianNoWinter10GSD_bgrn_split6_V74/Drop7-MedianNoWinter10GSD_bgrn_split6_V74_epoch46_step4042.pt'
                sc_pxl.package_fpath: 'models/fusion/Drop7-Cropped2GSD/packages/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84_epoch17_step1548.pt'
                bas_poly.thresh: 0.4
                bas_poly.time_thresh: 0.8
                sv_depth_filter.threshold: 0.1
                sv_dino_filter.end_min_score: 0.15
                sc_poly.smoothing: 0.66
                sc_poly.thresh: 0.1
                sc_poly.site_score_thresh: 0.35
                sc_crop.sensor_to_time_window: 'S2: 1month'
                '''
            ),
            (
                '''
                delivery: Eval17
                dag: KIT_TA2_PREEVAL17_BATCH_V126.py

                bas_pxl.package_fpath: 'models/fusion/uconn/D7-V2-COLD-candidate/epoch=203-step=4488.pt'
                bas_poly.thresh: 0.3875
                bas_poly.time_thresh: 0.8
                sv_dino_filter.end_min_score: 0.15
                sv_depth_score.model_fpath: models/depth_pcd/model4.h5
                sv_depth_filter.threshold: 0.1

                sc_pxl.package_fpath: 'models/fusion/Drop7-Cropped2GSD/packages/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84_epoch17_step1548.pt'
                sc_poly.smoothing: 0.66
                sc_poly.thresh: 0.1
                sc_poly.site_score_thresh: 0.35
                # sc_crop.sensor_to_time_window: 'S2: 1month'
                '''
            ),
            (
                '''
                delivery: Eval18
                dag: KIT_TA2_PREEVAL18_BATCH_V142.py

                bas_pxl.package_fpath: 'models/fusion/uconn/D7-V2-COLD-candidate/epoch=203-step=4488.pt'
                bas_poly.thresh: 0.3875
                bas_poly.time_thresh: 0.8
                sv_dino_filter.end_min_score: 0.15
                sv_depth_score.model_fpath: models/depth_pcd/model4.h5
                sv_depth_filter.threshold: 0.1

                sc_pxl.package_fpath: 'models/fusion/Drop7-Cropped2GSD/packages/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84_epoch17_step1548.pt'

                sc_poly.thresh: 0.3
                sc_poly.site_score_thresh: 0.3
                sc_poly.smoothing: 0.0
                sc_poly.boundaries_as: bounds
                sc_poly.new_algo: crall
                sc_poly.polygon_simplify_tolerance: 1
                # sc_crop.sensor_to_time_window: 'S2: 1month'
                '''
            ),

            (
                '''
                delivery: Eval19
                dag: KIT_TA2_PREEVAL18_BATCH_V138.py
                '''
            ),

            (
                '''
                delivery: Eval20
                dag: KIT_TA2_PREEVAL20_BATCH_V158.py
                '''
            ),
        ]

        for idx, text in enumerate(yamls_items):
            try:
                delivered_model_params += [Yaml.loads(text)]
            except Exception:
                print('ERROR')
                print(text)
                print(f'idx = {ub.urepr(idx, nl=1)}')
                raise

        ## SC
        # delivered_model_params += [
        #     {
        #         'name': 'Drop4_SC_RGB_scratch_V002_epoch=99-step=50300-v1.pt.pt',
        #         'sc_pxl.package_fpath': 'models/fusion/Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC/packages/Drop4_SC_RGB_scratch_V002/Drop4_SC_RGB_scratch_V002_epoch=99-step=50300-v1.pt.pt',
        #         'task': 'SC',
        #     },
        # ]
        return delivered_model_params

    def custom_channel_relabel(self, sub_macro_table, channel_key, coarsen=False):

        unique_channels = sub_macro_table['resolved_params.bas_pxl.channels'].unique()
        channel_maps = {
            'water|forest|field|impervious|barren|landcover_hidden:32': 'land',
            'blue|green|red|nir': 'raw' if coarsen else 'BGRN',
            'invariants:16': 'invar' if coarsen else 'invar1',
            'invariants:17': 'invar' if coarsen else 'invar1',
            'blue|green|red': 'raw' if coarsen else 'BGR',
            'red|green|blue': 'raw' if coarsen else 'BGR',
            'pan': 'raw' if coarsen else 'pan',
            'mae:16': 'mae',
            'sam:64': 'SAM',
            'mat_feats:16|materials:9|mtm': 'mat',
        }
        cold_maps = {}
        presentation_map = {}

        import kwcoco
        for c in unique_channels:
            if c is not None:
                new_streams = []
                sensorchan = kwcoco.SensorChanSpec.coerce(c)
                for stream in sensorchan.normalize().streams():
                    stream.sensor
                    concise_chans = stream.chans.concise().spec
                    if 'COLD' in concise_chans:
                        if coarsen:
                            cold_maps[concise_chans] = 'COLD'
                        else:
                            cold_maps[concise_chans] = 'COLD{}'.format(len(cold_maps) + 1)
                        channel_maps[concise_chans] = cold_maps[concise_chans]

                    if concise_chans in channel_maps:
                        concise_chans = channel_maps[concise_chans]
                    else:
                        print('WARNING: unregistered concise_chans = {}'.format(ub.urepr(concise_chans, nl=1)))
                    new_streams.append(f'{stream.sensor.spec}:{concise_chans}')
                new_sensorchan = kwcoco.SensorChanSpec.coerce(','.join(new_streams)).concise()

                if coarsen:
                    new_sensorchan = kwcoco.ChannelSpec.coerce(','.join(sorted(set(sc.chans.spec for sc in new_sensorchan.streams()))))
                    new_c = new_sensorchan.spec
                else:
                    new_c = new_sensorchan.spec
                presentation_map[c] = new_c

        new_columns = sub_macro_table['resolved_params.bas_pxl.channels'].apply(presentation_map.get)
        return new_columns

    def custom_channel_relabel_mapping(self, unique_channels, coarsen=False):
        channel_maps = {
            'water|forest|field|impervious|barren|landcover_hidden:32': 'land',
            'blue|green|red|nir': 'raw' if coarsen else 'BGRN',
            'invariants:16': 'invar' if coarsen else 'invar1',
            'invariants:17': 'invar' if coarsen else 'invar1',
            'blue|green|red': 'raw' if coarsen else 'BGR',
            'red|green|blue': 'raw' if coarsen else 'BGR',
            'pan': 'raw' if coarsen else 'pan',
            'mae:16': 'mae',
            'sam:64': 'SAM',
            'mat_feats:16|materials:9|mtm': 'mat',
        }
        cold_maps = {}
        presentation_map = {}

        import kwcoco
        for c in unique_channels:
            if c is not None:
                new_streams = []
                sensorchan = kwcoco.SensorChanSpec.coerce(c)
                for stream in sensorchan.normalize().streams():
                    stream.sensor
                    concise_chans = stream.chans.concise().spec
                    if 'COLD' in concise_chans:
                        if coarsen:
                            cold_maps[concise_chans] = 'COLD'
                        else:
                            # cold_maps[concise_chans] = 'COLD{}'.format(len(cold_maps) + 1)
                            cold_maps[concise_chans] = 'COLD.{}'.format(concise_chans.count('|') + 1)
                        channel_maps[concise_chans] = cold_maps[concise_chans]

                    if concise_chans in channel_maps:
                        concise_chans = channel_maps[concise_chans]
                    else:
                        print('WARNING: unregistered concise_chans = {}'.format(ub.urepr(concise_chans, nl=1)))

                    new_streams.append(f'{stream.sensor.spec}:{concise_chans}')
                new_sensorchan = kwcoco.SensorChanSpec.coerce(','.join(new_streams)).concise()

                if coarsen:
                    new_sensorchan = kwcoco.ChannelSpec.coerce(','.join(sorted(set(sc.chans.spec for sc in new_sensorchan.streams()))))
                    new_c = new_sensorchan.spec
                else:
                    new_c = new_sensorchan.spec
                presentation_map[c] = new_c
        return presentation_map

    def print_minmax_times(self, table):
        import pandas as pd
        from geowatch.utils.util_pandas import DotDictDataFrame
        from kwutil import util_time
        table = DotDictDataFrame(table)
        start_time_cols = table.search_columns('start_timestamp')
        end_time_cols = table.search_columns('stop_timestamp')
        timestamps = {}
        for k in start_time_cols + end_time_cols:
            timestamps[k] = table.loc[:, k].apply(lambda x: util_time.coerce_datetime(x) if not pd.isnull(x) else x)
        min_times = {}
        max_times = {}
        for k, vs in timestamps.items():
            try:
                min_times[k] = vs[~pd.isna(vs)].min()
                max_times[k] = vs[~pd.isna(vs)].max()
            except Exception as ex:
                print(f'ex={ex}')
        min_time = min(min_times.values())
        max_time = min(max_times.values())
        print(f'min_time={min_time.isoformat()}')
        print(f'max_time={max_time.isoformat()}')

    def threshold_param_groups(self, table, param_name, metric_name, metric_threshold):
        import pandas as pd
        param_groups = table.groupby(param_name)
        passed_thresh = param_groups[metric_name].describe()['max'] > metric_threshold
        filtered_table = pd.concat(list((ub.udict(list(param_groups)) & passed_thresh[passed_thresh].index).values()))
        return filtered_table


SMART_HELPER = SmartGlobalHelper()



import json
import os
import ubelt as ub
import scriptconfig as scfg


class SmartflowIngressConfig(scfg.DataConfig):
    """
    Ingress KWCOCO data to T&E baseline framework structure
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    assets = scfg.Value(None, type=str, position=2, required=True, help='Names of assets to download', nargs='+')
    outdir = scfg.Value(None, type=str, group='optional arguments', required=True, short_alias=['o'], help=ub.paragraph(
            '''
            Output directory for ingressed assets an output STAC Catalog
            '''))
    aws_profile = scfg.Value(None, type=str, group='optional arguments', help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands. UNUSED. Hook up to fsspec if needed.
            '''))
    dryrun = scfg.Value(False, isflag=True, group='optional arguments', short_alias=['d'], help='UNUSED. DEPRECATED')
    show_progress = scfg.Value(False, isflag=True, group='optional arguments', short_alias=['s'], help='UNUSED. DEPRECATED')
    dont_error_on_missing_asset = scfg.Value(False, isflag=True, group='optional arguments', help=ub.paragraph(
            '''
            Don't raise error on missing asset, just warn
            '''))


def main():
    smartflow_ingress(**SmartflowIngressConfig.cli())


def smartflow_ingress(input_path,
                      assets,
                      outdir,
                      aws_profile=None,
                      dryrun=False,
                      show_progress=False,
                      dont_error_on_missing_asset=False):
    """
    Downloads a STAC manifest and select items within it.

    Args:
        input_path (str):
            The path in the s3 bucket that the STAC item will be downloaded from.

        assets (List[str | Dict]):
            A List of keys into the stac item assets that we will download.
            Can also be a list of dictionaries that must contain a
            ``"key": <str>`` item, as well as other options to control
            behavior, like ``"allow_missing": True``.

        outdir (str | PathLike):
            local path to download to.

        aws_profile (str | None): aws cp argument

        dryrun (bool): aws cp argument

        show_progress (bool): aws cp argument

        dont_error_on_missing_asset (bool):
            if True warn if an asset is missing.
            TODO: variable name is too long and has a double negative.
            maybe rename to "missing_policy" or "ignore_missing"

    Returns:
        Dict[str, str | PathLike]:
            mapping from downloaded assets to their local path

    Example:
        >>> from geowatch.cli.smartflow_ingress import *  # NOQA
        >>> dpath = ub.Path.appdir('geowatch/tests/smartflow_ingress/dst').ensuredir()
        >>> fake_remote = ub.Path.appdir('geowatch/tests/smartflow_ingress/fake_remote').ensuredir()
        >>> fake_fpath = fake_remote / 'my_path.txt'
        >>> fake_fpath.write_text('foobar')
        >>> fake_dpath = (fake_remote / 'my_dir').ensuredir()
        >>> (fake_dpath / 'content1').touch()
        >>> (fake_dpath / 'content2').touch()
        >>> (fake_dpath / 'subdir1').ensuredir()
        >>> (fake_dpath / 'subdir1/subcontent1').touch()
        >>> (fake_dpath / 'subdir1/subcontent2').touch()
        >>> (fake_dpath / 'subdir1/subsubdir').ensuredir()
        >>> # Save this dummy stac item locally
        >>> # In practice we download it, but we are using dry run mode
        >>> # so we cant do that here.
        >>> demo_stac_content = {'raw_images': [],
        >>>  'stac': {'type': 'FeatureCollection',
        >>>   'features': [{'type': 'Feature',
        >>>     'stac_version': '1.0.0',
        >>>     'stac_extensions': [],
        >>>     'id': '66d3e2f605a44aa8b7bacc6ce7e96b9a',
        >>>     'geometry': {'type': 'Polygon',
        >>>      'coordinates': (((-109.56, 44.56),
        >>>        (-109.57, 44.55),
        >>>        (-109.53, 44.56),
        >>>        (-109.56, 44.56)),)},
        >>>     'bbox': [-109.57, 44.52, -109.51, 44.56],
        >>>     'properties': {},
        >>>     'assets': {'asset_file1': {'href': str(fake_fpath)},
        >>>      'asset_dir1': {'href': str(fake_dpath)}}}]}}
        >>> remote_dpath = (dpath / 'remote').ensuredir()
        >>> input_path = remote_dpath / 'items.jsonl'
        >>> input_path.write_text(json.dumps(demo_stac_content))
        >>> outdir = (dpath / 'local').ensuredir()
        >>> assets = ['asset_file1', 'asset_dir1', {'key': 'foobar', 'allow_missing': True}]
        >>> kwcoco_stac_item_assets = smartflow_ingress(
        >>>     input_path,
        >>>     assets,
        >>>     outdir,
        >>> )
        >>> assert kwcoco_stac_item_assets['asset_file1'] == os.fspath(outdir / 'my_path.txt')
        >>> assert kwcoco_stac_item_assets['asset_dir1'] == os.fspath(outdir / 'my_dir')
        >>> assert len(ub.Path(kwcoco_stac_item_assets['asset_dir1']).ls()) > 0
        >>> assert ub.Path(kwcoco_stac_item_assets['asset_file1']).exists()
    """
    print('--- BEGIN INGRESS ---')
    os.makedirs(outdir, exist_ok=True)

    assert aws_profile is None, 'unhandled'
    from geowatch.utils.util_fsspec import FSPath
    input_path = FSPath.coerce(input_path)

    def _loads_input(text):
        try:
            input_json = json.loads(text)
            return input_json['stac'].get('features', [])
        # Excepting KeyError here in case of a single line STAC item input
        except (json.decoder.JSONDecodeError, KeyError):
            # Support for simple newline separated STAC items
            lines = text.split('\n')
            return [json.loads(line) for line in lines]

    input_text = input_path.read_text()
    input_stac_items = _loads_input(input_text)

    # Our baseline KWCOCO egress script should only ever write out a
    # single KWCOCO STAC item
    if len(input_stac_items) != 1:
        raise RuntimeError("Expecting one and only one STAC item from input")
    kwcoco_stac_item = input_stac_items[0]
    kwcoco_stac_item_assets = {
        k: v['href'] for k, v in kwcoco_stac_item['assets'].items()}

    # TODO: can use fsspec to handle multiple downloads in parallel.
    seen = set()  # Prevent duplicate downloads
    for _asset in assets:
        if isinstance(_asset, str):
            asset_info = {'key': _asset}
        else:
            asset_info = _asset
        asset_key = asset_info['key']
        allow_missing = asset_info.get('allow_missing', dont_error_on_missing_asset)
        missing_action = asset_info.get('missing_action', 'ignore' if allow_missing else 'raise')

        try:
            asset_href = kwcoco_stac_item_assets[asset_key]
        except KeyError:
            missing_asset_str = (
                f"Expecting asset named {asset_key!r} in input KWCOCO STAC item"
            )
            if allow_missing:
                print(f"* Warning: {missing_asset_str!r}")
                continue
            else:
                raise RuntimeError(missing_asset_str)  # noqa

        outdir = FSPath.coerce(outdir)
        asset_href = FSPath.coerce(asset_href)
        asset_outpath = outdir / asset_href.name
        if asset_outpath not in seen:
            try:
                asset_href.copy(asset_outpath)
            except FileNotFoundError:
                missing_asset_str = (
                    f"Missing file for asset named {asset_key!r}"
                )
                if missing_action == 'ignore':
                    print(f"* Warning: {missing_asset_str!r}")
                    continue
                elif missing_action == 'mkdir':
                    ub.Path(str(asset_outpath)).ensuredir()
                elif missing_action == 'raise':
                    print('!!!')
                    print('!!!')
                    print('ERROR: occured while trying to ingress asset')
                    print('Printing debug information and then re-raising')
                    print('asset_href = {}'.format(ub.urepr(asset_href, nl=1)))
                    print('asset_outpath = {}'.format(ub.urepr(asset_outpath, nl=1)))
                    raise

        seen.add(asset_outpath)

        kwcoco_stac_item_assets[asset_key] = str(asset_outpath)

    # Returns assets (with downloaded asset hrefs updated)
    print('INGRESSED = {}'.format(ub.urepr(kwcoco_stac_item_assets, nl=1)))
    print('--- FINISH INGRESS ---')
    return kwcoco_stac_item_assets


if __name__ == "__main__":
    main()



import torch


def bilinear_interpolate(im, pos, batch=None):
    # From https://gist.github.com/peteflorence/a1da2c759ca1ac2b74af9a83f69ce20e
    x = pos[:, 1]
    y = pos[:, 0]
    x0 = torch.floor(x).long()
    x1 = x0 + 1

    y0 = torch.floor(y).long()
    y1 = y0 + 1

    x0_int = torch.clamp(x0, 0, im.shape[-1] - 1)
    x1_int = torch.clamp(x1, 0, im.shape[-1] - 1)
    y0_int = torch.clamp(y0, 0, im.shape[-2] - 1)
    y1_int = torch.clamp(y1, 0, im.shape[-2] - 1)

    if batch is not None:
        Ia = im[batch, :, y0_int, x0_int]
        Ib = im[batch, :, y1_int, x0_int]
        Ic = im[batch, :, y0_int, x1_int]
        Id = im[batch, :, y1_int, x1_int]
    else:
        Ia = im[..., y0_int, x0_int].t()
        Ib = im[..., y1_int, x0_int].t()
        Ic = im[..., y0_int, x1_int].t()
        Id = im[..., y1_int, x1_int].t()

    wa = (x1.float() - x) * (y1.float() - y)
    wb = (x1.float() - x) * (y - y0.float())
    wc = (x - x0.float()) * (y1.float() - y)
    wd = (x - x0.float()) * (y - y0.float())

    wa = wa.unsqueeze(1)
    wb = wb.unsqueeze(1)
    wc = wc.unsqueeze(1)
    wd = wd.unsqueeze(1)

    out = wa * Ia + wb * Ib + wc * Ic + wd * Id

    return out


def main():
    im = torch.tensor([
        [
            [0, 0.5, 0, 0],
            [0.25, 1, 0, 0],
        ],
        [
            [1, 1, 1, 1],
            [1, 1, 1, 1],
        ],
        [
            [2, 2, 2, 2],
            [2, 2, 2, 2],
        ]
    ], dtype=torch.float)
    im = im[:, None, ...]
    print(im.shape)
    print(im)
    pos = torch.tensor([
        [1.0, 0],
        [0.5, 0.5],
        [0.5, 0.5],
    ])
    batch = torch.tensor([0, 1, 2], dtype=torch.long)

    val = bilinear_interpolate(im, pos, batch=batch)
    print(val)


if __name__ == '__main__':
    main()
